from logging import Logger
from typing import Literal, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import CM_AS_INCH, TIME_VAR
from ....utils.read.product.file_info.type import FileType
from ....utils.read.product.level1.atl_nom_1b import get_depol_profile
from ....utils.time import TimedeltaLike, TimeRangeLike
from ....utils.typing import DistanceRangeLike
from ....utils.xarray_utils import filter_radius, filter_time
from ...color.colormap import get_cmap
from ...figure import (
    CurtainFigure,
    ECKFigure,
    FigureType,
    MapFigure,
    ProfileFigure,
    create_fig_layout_map_main_zoom_profile,
)
from .._cli import print_progress
from .._quicklook_results import _QuicklookResults


def ecquicklook_acth(
    ds: xr.Dataset,
    ds_bg: xr.Dataset,
    var_bg: str | None = None,
    vars: list[str] | None = None,
    show_maps: bool = True,
    show_zoom: bool = False,
    show_profile: bool = False,
    site: str | None = None,
    radius_km: float = 100.0,
    time_range: TimeRangeLike | None = None,
    height_range: DistanceRangeLike | None = (0, 30e3),
    ds_tropopause: xr.Dataset | None = None,
    ds_elevation: xr.Dataset | None = None,
    ds_temperature: xr.Dataset | None = None,
    resolution: Literal["low", "medium", "high", "l", "m", "h"] = "medium",
    closest_profile: bool = True,
    logger: Logger | None = None,
    log_msg_prefix: str = "",
    selection_max_time_margin: TimedeltaLike | Sequence[TimedeltaLike] | None = None,
    mode: Literal["fast", "exact"] = "fast",
) -> _QuicklookResults:
    _stime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    if not isinstance(ds_bg, xr.Dataset):
        raise TypeError(
            f"Invalid type '{ds_bg}' for dataset of background curtain (ds_bg)"
        )

    res: str
    if resolution.lower() in ["low", "l"]:
        res = "_low_resolution"
    elif resolution.lower() in ["medium", "m"]:
        res = "_medium_resolution"
    elif resolution.lower() in ["high", "h"]:
        res = ""
    else:
        raise ValueError(
            f'invalid resolution "{resolution}". valid values are: "low" or "l", "medium" or "m", "high" or "h".'
        )

    if not isinstance(var_bg, str):
        file_type_bg = FileType.from_input(ds_bg)
        if file_type_bg == FileType.ATL_NOM_1B:
            var_bg = "mie_attenuated_backscatter"
        elif file_type_bg == FileType.ATL_EBD_2A:
            var_bg = f"particle_backscatter_coefficient_355nm{res}"
        elif file_type_bg == FileType.ATL_AER_2A:
            var_bg = "particle_backscatter_coefficient_355nm"
        elif file_type_bg == FileType.ATL_TC__2A:
            var_bg = f"classification{res}"

    assert isinstance(var_bg, str)

    if vars is None:
        vars = [
            f"ATLID_cloud_top_height",
        ]

    wspaces: list[float] = []

    if show_maps:
        map_rows = [FigureType.MAP_1_ROW, FigureType.MAP_1_ROW]
        wspaces.append(3.5 * CM_AS_INCH)
    else:
        map_rows = None

    main_rows = [FigureType.CURTAIN for _ in vars]

    if show_zoom:
        zoom_rows = [FigureType.CURTAIN for _ in vars]
        wspaces.append(4 * CM_AS_INCH)
    else:
        zoom_rows = None

    profile_rows = None

    if logger:
        print_progress(f"layout", log_msg_prefix=log_msg_prefix, logger=logger)

    output = create_fig_layout_map_main_zoom_profile(
        map_rows=map_rows,
        main_rows=main_rows,
        zoom_rows=zoom_rows,
        profile_rows=profile_rows,
        wspace=wspaces,
    )
    fig, axs_map, axs_main, axs_zoom, axs_profile = output

    map_figs: list[ECKFigure] = []
    main_figs: list[ECKFigure] = []
    zoom_figs: list[ECKFigure] = []
    profile_figs: list[ECKFigure] = []

    if show_maps:
        if logger:
            print_progress(f"map globe", log_msg_prefix=log_msg_prefix, logger=logger)
        mf = MapFigure(ax=axs_map[0])
        mf = mf.ecplot(
            ds,
            site=site,
            radius_km=radius_km,
            time_range=time_range,
            selection_max_time_margin=selection_max_time_margin,
        )
        map_figs.append(mf)

        if logger:
            print_progress(f"map zoomed", log_msg_prefix=log_msg_prefix, logger=logger)
        mf = MapFigure(
            ax=axs_map[1],
            style="satellite",
            show_night_shade=False,
            show_right_labels=False,
            show_top_labels=False,
        )
        mf = mf.ecplot(
            ds,
            site=site,
            radius_km=radius_km,
            time_range=time_range,
            view="overpass",
            selection_max_time_margin=selection_max_time_margin,
        )
        map_figs.append(mf)

    for i, var in enumerate(vars):
        if logger:
            print_progress(
                f"curtain: var='{var_bg}'", log_msg_prefix=log_msg_prefix, logger=logger
            )
        cf = CurtainFigure(
            ax=axs_main[i],
            mode=mode,
        )
        cf = cf.ecplot(
            ds_bg,
            var_bg,
            site=site,
            radius_km=radius_km,
            selection_time_range=time_range,
            height_range=height_range,
            mark_closest_profile=closest_profile,
            cmap=get_cmap("calipso").blend(0.6),
            selection_max_time_margin=selection_max_time_margin,
        )
        if ds_tropopause:
            if logger:
                print_progress(
                    f"tropopause", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_tropopause(ds_tropopause)
        if ds_elevation:
            if logger:
                print_progress(
                    f"elevation", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_elevation(ds_elevation)
        if ds_temperature:
            if logger:
                print_progress(
                    f"temperature", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_temperature(ds_temperature)
        if var == "ATLID_cloud_top_height":
            if logger:
                print_progress(
                    f"line: {var=}",
                    log_msg_prefix=log_msg_prefix,
                    logger=logger,
                )
            cf = cf.ecplot_height(ds, var)

        main_figs.append(cf)

    if show_zoom or show_profile:
        _ds_bg = ds_bg.copy()
        if site:
            _ds_bg = filter_radius(_ds_bg, radius_km=radius_km, site=site)
        elif time_range:
            _ds_bg = filter_time(_ds_bg, time_range)

        if show_zoom:
            _time_range = _ds_bg[TIME_VAR].values[[0, -1]]
            for i, var in enumerate(vars):
                if logger:
                    print_progress(
                        f"curtain zoomed: var='{var_bg}'",
                        log_msg_prefix=log_msg_prefix,
                        logger=logger,
                    )
                cf = CurtainFigure(
                    ax=axs_main[i],
                    mode=mode,
                )
                cf = cf.ecplot(
                    ds_bg,
                    var_bg,
                    site=site,
                    radius_km=radius_km,
                    selection_time_range=time_range,
                    time_range=_time_range,
                    height_range=height_range,
                    mark_closest_profile=closest_profile,
                    cmap=get_cmap("calipso").blend(0.6),
                )
                if ds_tropopause:
                    if logger:
                        print_progress(
                            f"tropopause zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_tropopause(ds_tropopause)
                if ds_elevation:
                    if logger:
                        print_progress(
                            f"elevation zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_elevation(ds_elevation)
                if ds_temperature:
                    if logger:
                        print_progress(
                            f"temperature zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_temperature(ds_temperature)
                if var == "ATLID_cloud_top_height":
                    if logger:
                        print_progress(
                            f"line zoomed: {var=}",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_height(ds, var)

                zoom_figs.append(cf)

    subfigs: list[list[ECKFigure]] = [map_figs, main_figs, zoom_figs]  # , profile_figs]

    _etime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
    _dtime: str = str(pd.Timestamp(_etime) - pd.Timestamp(_stime)).split()[-1]
    if logger:
        print_progress(
            f"Plot created (time taken {_dtime}).",
            is_last=True,
            log_msg_prefix=log_msg_prefix,
            logger=logger,
        )

    return _QuicklookResults(fig, subfigs)
    return _QuicklookResults(fig, subfigs)



from logging import Logger
from typing import Literal, Sequence

import xarray as xr

from ...utils.constants import DEFAULT_PROFILE_SHOW_STEPS
from ...utils.ground_sites import GroundSite
from ...utils.read.product._generic import read_product
from ...utils.read.product.auxiliary import rebin_xmet_to_vertical_track
from ...utils.read.product.file_info.type import FileType
from ...utils.time import TimedeltaLike, TimeRangeLike
from ...utils.typing import DistanceRangeLike
from ..figure import ECKFigure
from ._level1 import ecquicklook_anom
from ._level2a import (
    ecquicklook_aaer,
    ecquicklook_acth,
    ecquicklook_aebd,
    ecquicklook_atc,
)
from ._quicklook_results import _QuicklookResults


def _get_addon_ds(
    ds: xr.Dataset,
    ds_filepath: str | None,
    ds_tropopause: xr.Dataset | str | None,
    ds_elevation: xr.Dataset | str | None,
    ds_temperature: xr.Dataset | str | None,
) -> tuple[xr.Dataset | None, xr.Dataset | None, xr.Dataset | None]:
    if (
        isinstance(ds_filepath, str)
        and isinstance(ds_tropopause, str)
        and ds_filepath == ds_tropopause
    ):
        ds_tropopause = ds

    if (
        isinstance(ds_filepath, str)
        and isinstance(ds_elevation, str)
        and ds_filepath == ds_elevation
    ):
        ds_elevation = ds

    if (
        isinstance(ds_filepath, str)
        and isinstance(ds_temperature, str)
        and ds_filepath == ds_temperature
    ):
        ds_temperature = ds

    if (
        isinstance(ds_tropopause, str)
        and isinstance(ds_elevation, str)
        and ds_tropopause == ds_elevation
    ):
        ds_elevation = ds_tropopause

    if (
        isinstance(ds_tropopause, str)
        and isinstance(ds_temperature, str)
        and ds_tropopause == ds_temperature
    ):
        ds_temperature = ds_tropopause

    if (
        isinstance(ds_elevation, str)
        and isinstance(ds_temperature, str)
        and ds_elevation == ds_temperature
    ):
        ds_temperature = ds_elevation

    if isinstance(ds_tropopause, str):
        ds_tropopause = read_product(ds_tropopause, in_memory=True)
    if isinstance(ds_elevation, str):
        ds_elevation = read_product(ds_elevation, in_memory=True)
    if isinstance(ds_temperature, str):
        ds_temperature = read_product(ds_temperature, in_memory=True)

    return ds_tropopause, ds_elevation, ds_temperature


def ecquicklook(
    ds: xr.Dataset | str,
    vars: str | list[str] | None = None,
    show_maps: bool = True,
    show_zoom: bool = False,
    show_profile: bool = True,
    site: GroundSite | str | None = None,
    radius_km: float = 100.0,
    time_range: TimeRangeLike | None = None,
    height_range: DistanceRangeLike | None = (0, 30e3),
    ds_tropopause: xr.Dataset | str | None = None,
    ds_elevation: xr.Dataset | str | None = None,
    ds_temperature: xr.Dataset | str | None = None,
    resolution: Literal["low", "medium", "high", "l", "m", "h"] = "low",
    ds2: xr.Dataset | str | None = None,
    ds_xmet: xr.Dataset | str | None = None,
    logger: Logger | None = None,
    log_msg_prefix: str = "",
    selection_max_time_margin: TimedeltaLike | Sequence[TimedeltaLike] | None = None,
    show_steps: bool = DEFAULT_PROFILE_SHOW_STEPS,
    mode: Literal["fast", "exact"] = "fast",
) -> _QuicklookResults:
    """
    Generate a preview visualization of an EarthCARE dataset with optional maps, zoomed views, and profiles.

    Args:
        ds (xr.Dataset | str): EarthCARE dataset or path.
        vars (str | list[str] | None, otional): List of variable to plot. Automatically sets product-specific default list of variables if None.
        show_maps (bool, optional): Whether to include map view. Dafaults to True.
        show_zoom (bool, optional): Whether to show an additional column of zoomed plots. Defaults to False.
        show_profile (bool, optional): Whether to include vertical profile plots. Dfaults to True.
        site (GroundSite | str | None, optional): Ground site object or name identifier.
        radius_km (float, optional): Search radius around site in kilometers. Defaults to 100.
        time_range (TimeRangeLike | None, optional): Time range filter.
        height_range (DistanceRangeLike | None, optional): Height range in meters. Defaults to (0, 30_000).
        ds_tropopause (xr.Dataset | str | None, optional): Optional dataset or path containing tropopause data to add it to the plot.
        ds_elevation (xr.Dataset | str | None, optional): Optional dataset or path containing elevation data to add it to the plot.
        ds_temperature (xr.Dataset | str | None, optional): Optional dataset or path containing temperature data to add it to the plot.
        resolution (Literal["low", "medium", "high", "l", "m", "h"], optional): Resolution of A-PRO data. Defaults to "low".
        ds2 (xr.Dataset | str | None, optional): Secondary dataset required for certain product quicklook (e.g., A-LAY products need A-NOM or A-EBD to serve as background curtain plots).
        ds_xmet(xr.Dataset | str | None, optional): Optional auxiliary meteorological dataset used to plot tropopause, elevation and temperature from.
        logger (Logger, optional): Logger instance for output messages.
        log_msg_prefix (str, optional): Prefix for log messages.
        selection_max_time_margin (TimedeltaLike | Sequence[TimedeltaLike] | None, optional): Allowed time difference for selection.
        show_steps (bool, optional): Whether to plot profiles as height bin step functions or instead plot only the line through bin centers. Defaults to True.
        mode (Literal["fast", "exact"], optional): Processing mode.

    Returns:
        _QuicklookResults: Object containing figures and metadata.
    """
    if isinstance(vars, str):
        vars = [vars]

    filepath: str | None = None
    if isinstance(ds, str):
        filepath = ds

    ds = read_product(ds, in_memory=True)
    file_type = FileType.from_input(ds)

    if isinstance(ds_xmet, (xr.Dataset, str)):
        ds_xmet = read_product(ds_xmet, in_memory=True)
        if file_type in [
            FileType.ATL_NOM_1B,
            FileType.ATL_FM__2A,
            FileType.ATL_AER_2A,
            FileType.ATL_EBD_2A,
            FileType.ATL_ICE_2A,
            FileType.ATL_TC__2A,
            FileType.ATL_CLA_2A,
            FileType.CPR_NOM_1B,
        ]:
            ds_xmet = rebin_xmet_to_vertical_track(ds_xmet, ds)

    ds_tropopause, ds_elevation, ds_temperature = _get_addon_ds(
        ds,
        filepath,
        ds_tropopause or ds_xmet,
        ds_elevation or ds_xmet,
        ds_temperature or ds_xmet,
    )

    kwargs = dict(
        ds=ds,
        vars=vars,
        show_maps=show_maps,
        show_zoom=show_zoom,
        show_profile=show_profile,
        site=site,
        radius_km=radius_km,
        time_range=time_range,
        height_range=height_range,
        ds_tropopause=ds_tropopause,
        ds_elevation=ds_elevation,
        ds_temperature=ds_temperature,
        logger=logger,
        log_msg_prefix=log_msg_prefix,
        selection_max_time_margin=selection_max_time_margin,
        mode=mode,
    )

    if file_type == FileType.ATL_NOM_1B:
        kwargs["show_steps"] = show_steps
        return ecquicklook_anom(**kwargs)  # type: ignore
    elif file_type == FileType.ATL_EBD_2A:
        kwargs["show_steps"] = show_steps
        kwargs["resolution"] = resolution
        return ecquicklook_aebd(**kwargs)  # type: ignore
    elif file_type == FileType.ATL_AER_2A:
        kwargs["show_steps"] = show_steps
        kwargs["resolution"] = resolution
        return ecquicklook_aaer(**kwargs)  # type: ignore
    elif file_type == FileType.ATL_TC__2A:
        return ecquicklook_atc(**kwargs)  # type: ignore
    elif file_type == FileType.ATL_CTH_2A:

        if ds2 is not None:
            ds2 = read_product(ds2, in_memory=True)
            file_type2 = FileType.from_input(ds2)
            if file_type2 in [
                FileType.ATL_NOM_1B,
                FileType.ATL_EBD_2A,
                FileType.ATL_AER_2A,
                FileType.ATL_TC__2A,
            ]:
                kwargs["ds_bg"] = ds2
                kwargs["resolution"] = resolution
                return ecquicklook_acth(**kwargs)  # type: ignore
            raise ValueError(
                f"There is no CTH background curtain plotting for {str(file_type2)} products. Use instead: {str(FileType.ATL_NOM_1B)}, {str(FileType.ATL_EBD_2A)}, {str(FileType.ATL_AER_2A)}, {str(FileType.ATL_TC__2A)}"
            )
        raise TypeError(f"""Missing dataset "ds2" to plot a background for the CTH""")

    raise NotImplementedError()



"""
earthcarekit
============

A Python package to simplify working with EarthCARE satellite data.

Copyright (c) 2025 Leonard König

Licensed under the MIT License (see LICENSE file or https://opensource.org/license/mit).
"""

__author__ = "Leonard König"
__license__ = "MIT"
__version__ = "0.2.0"
__date__ = "2025-08-12"
__maintainer__ = "Leonard König"
__email__ = "koenig@tropos.de"
__title__ = "earthcarekit"

from .calval import *
from .download import ecdownload
from .plot import *
from .plot import ecquicklook, ecswath
from .utils import ProfileData, filter_radius, filter_time, set_config
from .utils import statistics as stats
from .utils.config import _warn_user_if_not_default_config_exists, create_example_config
from .utils.geo import geodesic, haversine
from .utils.ground_sites import GroundSite, get_ground_site
from .utils.logging import _setup_logging
from .utils.overpass import get_overpass_info
from .utils.read import *

__all__ = [
    "ecquicklook",
    "ecswath",
    "ecdownload",
]

_setup_logging()
_warn_user_if_not_default_config_exists()



from ._types import _RadiusSearch


def parse_radius_search(
    radius_search: list[str | float] | None,
) -> _RadiusSearch:
    """Convert user's radius inputs to query parameters, that can be used in search requests."""
    if radius_search is not None:
        radius = str(int(radius_search[0]))
        lat = str(float(radius_search[1]))
        lon = str(float(radius_search[2]))
        return _RadiusSearch(radius=radius, lat=lat, lon=lon)
    return _RadiusSearch(radius=None, lat=None, lon=None)



from typing import Final

URL_OADS: Final[str] = "https://eocat.esa.int/eo-catalogue"
URL_MAAP: Final[str] = "https://catalog.maap.eo.esa.int/catalogue"
URL_ENTRYPOINT: Final[str] = URL_MAAP
URL_AUTHENTIFICATION: Final[str] = "https://eoiam-idp.eo.esa.int/samlsso"

PROGRAM_NAME: Final[str] = "ecdownload"
PROGRAM_DESCRIPTION: Final[str] = """description is missing"""
PROGRAM_SETUP_INSTRUCTIONS: Final[str] = """setup instructions are missing"""



import os
from typing import Final

TEMPORARY_DIRNAME: Final[str] = "tmp"


def get_tmp_dirpath() -> str:
    """Returns absoulute path to the folder where temporary files produces by `earthcare-kit` are stored."""
    script_filepath = os.path.abspath(__file__)
    script_dirpath = os.path.dirname(script_filepath)
    tmp_dirpath = os.path.abspath(
        os.path.join(script_dirpath, os.pardir, os.pardir, os.pardir, TEMPORARY_DIRNAME)
    )
    print(tmp_dirpath)
    return tmp_dirpath



import warnings

import numpy as np
import xarray as xr
from scipy.interpolate import griddata  # type: ignore

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    ELEVATION_VAR,
    HEIGHT_VAR,
)
from ....rolling_mean import rolling_mean_2d
from ....xarray_utils import filter_time, merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def get_depol_profile(
    ds: xr.Dataset,
    cpol_cleaned_var: str = "cpol_cleaned_for_depol_calculation",
    xpol_cleaned_var: str = "xpol_cleaned_for_depol_calculation",
):
    cpol = ds[cpol_cleaned_var].values
    xpol = ds[xpol_cleaned_var].values
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        mean_xpol_bsc = np.nanmean(xpol, axis=0)
        mean_mie_bsc = np.nanmean(cpol, axis=0)
    return mean_xpol_bsc / mean_mie_bsc


def add_depol_ratio(
    ds: xr.Dataset,
    rolling_w: int = 20,
    near_zero_tolerance: float = 2e-7,
    smooth: bool = True,
    skip_height_above_elevation: int = 3,
    depol_ratio_var: str = "depol_ratio",
    cpol_cleaned_var: str = "cpol_cleaned_for_depol_calculation",
    xpol_cleaned_var: str = "xpol_cleaned_for_depol_calculation",
    cpol_var: str = "mie_attenuated_backscatter",
    xpol_var: str = "crosspolar_attenuated_backscatter",
    elevation_var: str = ELEVATION_VAR,
    height_var: str = HEIGHT_VAR,
) -> xr.Dataset:
    cpol_da = ds[cpol_var].copy()
    xpol_da = ds[xpol_var].copy()
    ds[depol_ratio_var] = xpol_da / cpol_da
    rename_var_info(
        ds,
        depol_ratio_var,
        name=depol_ratio_var,
        long_name="Depol. ratio from cross- and co-polar atten. part. bsc.",
        units="",
    )

    elevation = (
        ds[elevation_var].values.copy()[:, np.newaxis] + skip_height_above_elevation
    )
    mask_surface = ds[height_var].values[0].copy() < elevation

    xpol = ds[xpol_var].values
    cpol = ds[cpol_var].values
    xpol[mask_surface] = np.nan
    cpol[mask_surface] = np.nan
    if smooth:
        xpol = rolling_mean_2d(xpol, rolling_w, axis=0)
        cpol = rolling_mean_2d(cpol, rolling_w, axis=0)
        near_zero_mask = np.isclose(cpol, 0, atol=near_zero_tolerance)
        ds[depol_ratio_var].values = xpol / cpol
        ds[depol_ratio_var].values[near_zero_mask] = np.nan
    else:
        ds[depol_ratio_var].values = xpol / cpol

    ds[cpol_cleaned_var] = ds[cpol_var].copy()
    ds[cpol_cleaned_var].values = cpol

    ds[xpol_cleaned_var] = ds[xpol_var].copy()
    ds[xpol_cleaned_var].values = xpol

    return ds


def read_product_anom(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens ATL_NOM_1B file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    # Since ATLID is angled backwards a time shift of nearly 3 seconds is created which is corrected here to
    ds["original_time"] = ds["time"].copy()
    ds["time"].values = ds["time"].values + np.timedelta64(-2989554432, "ns")

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="height",
        track_lat_var="ellipsoid_latitude",
        track_lon_var="ellipsoid_longitude",
        height_var="sample_altitude",
        time_var="time",
        temperature_var="layer_temperature",
        elevation_var="surface_elevation",
    )
    ds = rename_var_info(
        ds,
        "mie_attenuated_backscatter",
        "Co-polar atten. part. bsc.",
        "Co-polar atten. part. bsc.",
        "m$^{-1}$ sr$^{-1}$",
    )
    ds = rename_var_info(
        ds,
        "rayleigh_attenuated_backscatter",
        "Ray. atten. bsc.",
        "Ray. atten. bsc.",
        "m$^{-1}$ sr$^{-1}$",
    )
    ds = rename_var_info(
        ds,
        "crosspolar_attenuated_backscatter",
        "Cross-polar atten. part. bsc.",
        "Cross-polar atten. part. bsc.",
        "m$^{-1}$ sr$^{-1}$",
    )
    ds = add_depol_ratio(ds)

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



import re
from logging import Logger

from ._exceptions import InvalidInputError
from ._types import ProductTypeVersion


def parse_product_type_and_version(
    input_string: str,
    logger: Logger | None = None,
    default_version: str = "latest",
) -> ProductTypeVersion:
    """Returns a tuple of formatted product name and baseline strings (allows short names as input, e.g. 'ANOM:AA' -> ('ATL_NOM_1B', 'AA'))."""
    product_name_input = (
        input_string.replace(" ", "").replace("-", "").replace("_", "").lower()
    )
    product_baseline = None
    tmp = product_name_input.split(":")
    if len(tmp) == 2:
        product_baseline = tmp[1].upper()
        if re.fullmatch("[A-Z]{2}", product_baseline) is None:
            exception_msg = f"Product version in '{input_string}' is not valid. Please specify the product version by giving the two-letter processor baseline after the colon (e.g. ':AC')."
            if logger:
                logger.exception(exception_msg)
            raise InvalidInputError(exception_msg)
        product_name_input = tmp[0]

    file_types = [
        # ATLID level 1b
        "ATL_NOM_1B",
        "ATL_DCC_1B",
        "ATL_CSC_1B",
        "ATL_FSC_1B",
        # MSI level 1b
        "MSI_NOM_1B",
        "MSI_BBS_1B",
        "MSI_SD1_1B",
        "MSI_SD2_1B",
        # BBR level 1b
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "BBR_SOL_1B",
        "BBR_LIN_1B",
        # CPR level 1b
        "CPR_NOM_1B",  # JAXA product
        # MSI level 1c
        "MSI_RGR_1C",
        # level 1d
        "AUX_MET_1D",
        "AUX_JSG_1D",
        # ATLID level 2a
        "ATL_FM__2A",
        "ATL_AER_2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "ATL_EBD_2A",
        "ATL_CTH_2A",
        "ATL_ALD_2A",
        "ATL_CLA_2A",  # JAXA product
        # MSI level 2a
        "MSI_CM__2A",
        "MSI_COP_2A",
        "MSI_AOT_2A",
        "MSI_CLP_2A",  # JAXA product
        # CPR level 2a
        "CPR_FMR_2A",
        "CPR_CD__2A",
        "CPR_TC__2A",
        "CPR_CLD_2A",
        "CPR_APC_2A",
        "CPR_ECO_2A",  # JAXA product
        "CPR_CLP_2A",  # JAXA product
        # ATLID-MSI level 2b
        "AM__MO__2B",
        "AM__CTH_2B",
        "AM__ACD_2B",
        # ATLID-CPR level 2b
        "AC__TC__2B",
        "AC__CLP_2B",  # JAXA product
        # BBR-MSI-(ATLID) level 2b
        "BM__RAD_2B",
        "BMA_FLX_2B",
        # ATLID-CPR-MSI level 2b
        "ACM_CAP_2B",
        "ACM_COM_2B",
        "ACM_RT__2B",
        "ACM_CLP_2B",  # JAXA product
        # ATLID-CPR-MSI-BBR
        "ALL_DF__2B",
        "ALL_3D__2B",
        "ALL_RAD_2B",  # JAXA product
        # Orbit data
        "MPL_ORBSCT",  # Orbit scenario file
        "AUX_ORBPRE",  # Predicted orbit file
        "AUX_ORBRES",  # Restituted/reconstructed orbit file
    ]

    short_names = []

    for file_type in file_types:
        long_name = file_type.replace("_", "").lower()
        medium_name = long_name[0:-2]
        short_name = medium_name
        string_replacements = [
            ("atl", "a"),
            ("msi", "m"),
            ("bbr", "b"),
            ("cpr", "c"),
            ("aux", "x"),
        ]
        for old_string, new_string in string_replacements:
            short_name = short_name.replace(old_string, new_string)

        expected_inputs = [long_name, medium_name, short_name]

        if "ALL_" == file_type[0:4]:
            alternative_long_name = "acmb" + long_name[3:]
            alternative_short_name = "acmb" + short_name[3:]
            expected_inputs.extend([alternative_long_name, alternative_short_name])

        if product_name_input.lower() in expected_inputs:
            if product_baseline is not None:
                return ProductTypeVersion(type=file_type, version=product_baseline)
            else:
                return ProductTypeVersion(type=file_type, version=default_version)

        short_names.append(short_name.upper())

    msg = ""
    msg2 = ""
    for i in range(len(file_types)):
        if i % 6 == 0:
            msg += "\n" + file_types[i]
            msg2 += "\n" + short_names[i]
        else:
            msg += "\t" + file_types[i]
            msg2 += "\t" + short_names[i]

    exception_msg = f'The user input "{input_string}" is either not a valid product name or not supported by this function.\n{msg}\n\nor use the respective short hands (additional non letter characters like - or _ are also allowed, e.g. A-NOM):\n{msg2}'
    if logger:
        logger.exception(exception_msg)
    raise InvalidInputError(exception_msg)



import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex


def get_cmap():
    """
    Based on color index values from the chiljet2 colormap in the 'ectools' project by Shannon Mason (ECMWF):
    https://bitbucket.org/smason/workspace/projects/EC
    Original under Apache 2.0 license.
    """
    name = "chiljet2"
    colors = [
        "#e6e6e6",
        "#0000ff",
        "#008000",
        "#ffff00",
        "#ff0000",
        "#000000",
    ]
    positions = [0.0, 0.2, 0.35, 0.67, 0.9, 1.0]
    color_dict = list(zip(positions, colors))
    cmap = LinearSegmentedColormap.from_list(name, color_dict)
    return cmap



from .convertsions import geo_to_ecef, sequence_geo_to_ecef
from .coordinates import get_coords
from .distance import geodesic, get_cumulative_distances, haversine, vincenty
from .interpolate import interpgeo



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap():
    cmap_data = [
        (-3, "#F2F3DE", "Missing"),
        (-2, "#000000", "Surface"),
        (-1, "#696969", "Attenuated"),
        (0, "#FFFFFF", "Clear"),
        (1, "#0000FF", "(Warm) Liquid cloud"),
        (2, "#010089", "S'cooled cloud"),
        (3, "#00FFFF", "Ice cloud"),
        (10, "#A52A2C", "Dust"),
        (11, "#ADD8E6", "Sea salt"),
        (12, "#01FD80", "Continental pollution"),
        (13, "#2B4F52", "Smoke"),
        (14, "#CE853E", "Dusty smoke"),
        (15, "#DFB788", "Dusty mix"),
        (20, "#FF00FF", "STS"),
        (21, "#8929E1", "NAT"),
        (22, "#4B0080", "Strat. ice"),
        (25, "#FBE5B5", "Strat. ash"),
        (26, "#FEFF00", "Strat. sulfate"),
        (27, "#BDB66C", "Strat. smoke"),
        # (101, , 'Unknown: Aerosol Target has a very low probability (no class assigned)')
        # (102, , 'Unknown: Aerosol classification outside of param space')
        # (104, , 'Unknown: Strat. Aerosol Target has a very low probability (no class assigned)')
        # (105, , 'Unknown: Strat. Aerosol classification outside of param space')
        # (106, , 'Unknown: PSC Target has a very low probability (no class assigned)')
        # (107, , 'Unknown: PSC classification outside of param space')
    ]
    colors = [c for _, c, _ in cmap_data]
    definitions = {k: l for k, _, l in cmap_data}
    cmap = Cmap(colors=colors, name="target_classification").to_categorical(definitions)
    return cmap



from ._types import CollectionStr, ProductTypeStr, ProductTypeVersion


def get_collection_names_matching_product_availability(
    product: ProductTypeVersion,
    collection_product_type_dict: dict[CollectionStr, list[ProductTypeStr]],
) -> list[CollectionStr]:
    """Returns names of collections that contain the desired product according to the given `dict`."""
    return [k for k, v in collection_product_type_dict.items() if product.type in v]



import argparse
import datetime
import os
import sys
from argparse import RawTextHelpFormatter
from dataclasses import dataclass
from typing import Any, Final

import numpy as np
import pandas as pd
import xarray as xr

from ... import __title__, __version__
from ...utils import FileType, GroundSite, read_product, search_product
from ...utils._cli import (
    console_exclusive_info,
    create_logger,
    get_counter_message,
    log_textbox,
)
from ...utils._cli._parse import (
    parse_path_to_config,
    parse_path_to_data,
    parse_path_to_imgs,
    parse_search_inputs,
    parse_selected_index,
)
from ...utils._cli._parse._types import _SearchInputs
from ...utils.config import ECKConfig
from ..save import create_filepath, save_plot
from ._quicklook import ecquicklook

PROGRAM_NAME: Final[str] = "ecquicklook"
PROGRAM_DESCRIPTION: Final[str] = """description is missing"""
PROGRAM_SETUP_INSTRUCTIONS: Final[str] = """setup instructions are missing"""


def add_args_product_search(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        "-fp",
        "--filepath",
        type=str,
        nargs="*",
        default=None,
        help="A list of paths to EarthCARE product files (.h5 files).",
    )
    parser.add_argument(
        "-p",
        "--product_type",
        type=str,
        nargs="*",
        default=None,
        help="A list of EarthCARE product names (e.g. ANOM or ATL-NOM-1B, etc.).\nNote: You can also specify the product version by adding a colon and the two-letter\nprocessor baseline after the name (e.g. ANOM:AD).",
    )
    parser.add_argument(
        "-pv",
        "--product_version",
        type=str,
        default="latest",
        help='Product version, i.e. the two-letter identifier of the processor baseline (e.g. AC). Defaults to "latest"',
    )
    parser.add_argument(
        "-o",
        "--orbit_number",
        type=int,
        nargs="*",
        default=None,
        help="A list of EarthCARE orbit numbers (e.g. 981).\nNote: Can not be used with combined orbit and frame options -soaf and -eoaf.",
    )
    parser.add_argument(
        "-f",
        "--frame_id",
        type=str,
        nargs="*",
        default=None,
        help="A EarthCARE frame ID (i.e. single letters from A to H).\nNote: Can not be used with combined orbit and frame options -soaf and -eoaf.",
    )
    parser.add_argument(
        "-oaf",
        "--orbit_and_frame",
        type=str,
        nargs="*",
        default=None,
        help="A string describing the EarthCARE orbit number and frame (e.g. 00981E)",
    )
    parser.add_argument(
        "-t",
        "--time",
        type=str,
        nargs="*",
        default=None,
        help='Search for data containing a specific timestamp (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    return parser


def main() -> None:
    time_start_script: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    parser = argparse.ArgumentParser(
        prog=PROGRAM_NAME,
        description=f"{PROGRAM_DESCRIPTION}\n\n{PROGRAM_SETUP_INSTRUCTIONS}",
        formatter_class=RawTextHelpFormatter,
    )
    parser = add_args_product_search(parser)
    parser.add_argument(
        "-hmin",
        "--min_height",
        type=float,
        default=0.0,
        help="The minimum height plotted in kilometers. Defaults to 0.0",
    )
    parser.add_argument(
        "-hmax",
        "--max_height",
        type=float,
        default=30.0,
        help="The maximum height plotted in kilometers. Defaults to 30.0",
    )
    parser.add_argument(
        "-rad",
        "--site_radius",
        type=float,
        default=100.0,
        help="The radius around a ground site or geo location in kilometers. Defaults to 100.0",
    )
    parser.add_argument(
        "-geo",
        "--site_geo_location",
        type=float,
        nargs=2,
        default=None,
        help="The latitude and logitude coordniates of a ground site in degrees north and east.",
    )
    parser.add_argument(
        "-site",
        "--site_name",
        type=str,
        default=None,
        help="The name of a ground site.",
    )
    parser.add_argument(
        "-d",
        "--path_to_data",
        type=str,
        default=None,
        help="The local root directory where EarthCARE products are stored.",
    )
    parser.add_argument(
        "-img",
        "--image_directory",
        type=str,
        default=None,
        help="The local directory where producted images will be saved to.",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrites local data (otherwise existing files will not be processed again).",
    )
    parser.add_argument(
        "-c",
        "--path_to_config",
        type=str,
        default=None,
        help="The path to an OADS credential TOML file (note: if not provided, a file named 'config.toml' is required in the script's folder)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Shows debug messages in console."
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="Shows the tools's version and exits.",
    )
    args = parser.parse_args()

    if args.version:
        print(f"earthcarekit {__version__}")
        sys.exit(0)

    is_overwrite: bool = args.overwrite
    is_debug: bool = args.debug
    hmin: float = args.min_height * 1000.0  # km to m
    hmax: float = args.max_height * 1000.0  # km to m
    radius_km: float = args.site_radius
    site_lat: float | None = (
        None
        if not isinstance(args.site_geo_location, list)
        else args.site_geo_location[0]
    )
    site_lon: float | None = (
        None
        if not isinstance(args.site_geo_location, list)
        else args.site_geo_location[1]
    )
    site_name: str | None = args.site_name

    logger = create_logger(
        logger_name=PROGRAM_NAME,
        log_to_file=False,
        debug=is_debug,
    )

    log_textbox(
        f"EarthCARE Quicklook Tool\n{__title__} {__version__}",
        logger=logger,
        is_mayor=True,
    )

    config = parse_path_to_config(args.path_to_config, logger=logger)
    path_to_data = parse_path_to_data(args.path_to_data, logger=logger)
    if isinstance(path_to_data, str):
        config.path_to_data = path_to_data
    path_to_imgs = parse_path_to_imgs(args.image_directory, logger=logger)
    if isinstance(path_to_imgs, str):
        config.path_to_images = path_to_imgs

    logger.info(f"# Settings")
    logger.info(f"# - config_filepath=<{config.filepath}>")
    logger.info(f"# - data_directory=<{config.path_to_data}>")
    logger.info(f"# - image_directory=<{config.path_to_images}>")
    logger.info(f"# - height_range: {args.min_height} to {args.max_height} km")
    logger.info(f"# - {site_name=}")
    logger.info(f"# - {radius_km=}")
    logger.info(f"# - site_coords=({site_lat}, {site_lon})")
    logger.info(f"# - {is_debug=}")
    logger.info(f"# - {is_overwrite=}")
    console_exclusive_info()

    file_type: list[str]
    baseline: list[str]
    if args.product_type:
        search_inputs: _SearchInputs = parse_search_inputs(
            product_type=args.product_type,
            product_version=args.product_version,
            logger=logger,
        )
        file_type = [p.type for p in search_inputs.products]
        baseline = [p.version for p in search_inputs.products]
    else:
        file_type = []
        baseline = []

    df = search_product(
        root_dirpath=config.path_to_data,
        config=config,
        file_type=file_type,
        baseline=baseline,
        timestamp=args.time,
        orbit_and_frame=args.orbit_and_frame,
        orbit_number=args.orbit_number,
        frame_id=args.frame_id,
        filename=args.filepath,
    )

    num_plots: int = 0
    for i, (_, row) in enumerate(df.iterrows()):
        count_msg, _ = get_counter_message(counter=i + 1, total_count=len(df))

        logger.info(f"*{count_msg} {row['name']}")

        site: GroundSite | str | None = site_name
        _site_name: str | None = site_name
        if isinstance(site_lat, float) and isinstance(site_lon, float):
            site = GroundSite(
                latitude=site_lat,
                longitude=site_lon,
                name=site_name or "X",
                long_name=site_name or "X",
            )
            _lat_str = f"{site_lat}N" if site_lat >= 0 else f"{-site_lat}S"
            _lon_str = f"{site_lon}E" if site_lon >= 0 else f"{-site_lon}W"
            _site_name = site_name or f"{_lat_str}{_lon_str}"
            _site_name = _site_name.replace(".", "d")

        ds2: xr.Dataset | str | None = None
        if row["file_type"] == FileType.ATL_CTH_2A:
            for _bg_file_type in ["anom", "aebd"]:
                _df = search_product(
                    root_dirpath=config.path_to_data,
                    config=config,
                    file_type=_bg_file_type,
                    orbit_and_frame=row["orbit_and_frame"],
                )
                if len(_df) > 0:
                    ds2 = _df.filepath[-1]
            if ds2 is None:
                logger.info(
                    f" {count_msg} Skipping since no matching A-NOM or A-EBD frame was found to use as the curtain background"
                )
                continue

        with read_product(row["filepath"]) as ds:
            img_filepath = create_filepath(
                filename="quicklook.png",
                ds=ds,
                hmax=hmax,
                site_name=_site_name,
                radius=radius_km if site else None,
            )
            if os.path.exists(img_filepath) and not is_overwrite:
                logger.info(
                    f" {count_msg} Skipping since image already exits at <{img_filepath}>"
                )
                continue
            gl = ecquicklook(
                ds=ds,
                ds2=ds2,
                logger=logger,
                log_msg_prefix=f" {count_msg} ",
                height_range=(hmin, hmax),
                site=site,
                radius_km=radius_km,
            )
            save_plot(
                fig=gl.fig,
                filepath=img_filepath,
                verbose=True,
                print_prefix=f" {count_msg} ",
            )
            num_plots += 1

    time_end_script: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
    execution_time: pd.Timedelta = pd.Timestamp(time_end_script) - pd.Timestamp(
        time_start_script
    )
    execution_time_str = str(execution_time).split()[-1]

    console_exclusive_info()
    _msg = [
        f"EXECUTION SUMMARY",
        "---",
        f"Time taken          {execution_time_str}",
        f"Files found         {len(df)}",
        f"Quicklooks created  {num_plots}",
        # f"Errors occured      {num_errors}",
    ]
    log_textbox("\n".join(_msg), logger=logger, show_time=True)


if __name__ == "__main__":
    main()



import logging
from typing import Final

LOG_FORMAT_USER: Final[str] = "[%(levelname)s] - %(message)s"
LOG_FORMAT_DEV: Final[str] = (
    "%(asctime)s [%(levelname).1s] %(name)s.%(funcName)s:%(lineno)d - %(message)s"
)
LOG_FORMAT_JSON: Final[str] = (
    '{ "time": "%(asctime)s", "level": "%(levelname).1s", "module": "%(name)s", "message": "%(message)s" }'
)
LOG_FORMAT_LINE: Final[str] = (
    "%(asctime)s [%(levelname).1s] %(pathname)s:%(lineno)d - %(message)s"
)


def _setup_logging(
    level: int | str = logging.INFO, format: str = LOG_FORMAT_USER
) -> None:
    logging.getLogger("matplotlib").setLevel(logging.ERROR)
    logging.basicConfig(level=level, format=format)



import os

import numpy as np
from xarray import Dataset

from ...constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ...xarray_utils import merge_datasets
from ._trim_to_frame import trim_to_latitude_frame_bounds
from .auxiliary import read_product_xmet
from .file_info import FileType
from .header_group import add_header_and_meta_data
from .level1 import read_product_anom, read_product_cnom, read_product_mrgr
from .level2a import (
    read_product_aaer,
    read_product_aald,
    read_product_acla,
    read_product_acth,
    read_product_aebd,
    read_product_atc,
)
from .level2b import read_product_amacd, read_product_amcth


def _read_auxiliary_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta]
    if file_type == FileType.AUX_MET_1D:
        return read_product_xmet(*args)
    else:
        return None


def _read_level1_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta]
    if file_type == FileType.ATL_NOM_1B:
        return read_product_anom(*args)
    elif file_type == FileType.MSI_RGR_1C:
        return read_product_mrgr(*args)
    elif file_type == FileType.CPR_NOM_1B:
        return read_product_cnom(*args)
    else:
        return None


def _read_level2a_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta]
    if file_type == FileType.ATL_AER_2A:
        return read_product_aaer(*args)
    elif file_type == FileType.ATL_EBD_2A:
        return read_product_aebd(*args)
    elif file_type == FileType.ATL_TC__2A:
        return read_product_atc(*args)
    elif file_type == FileType.ATL_CLA_2A:
        return read_product_acla(*args)
    elif file_type == FileType.ATL_CTH_2A:
        return read_product_acth(*args)
    elif file_type == FileType.ATL_ALD_2A:
        return read_product_aald(*args)
    else:
        return None


def _read_level2b_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta]
    if file_type == FileType.AM__ACD_2B:
        return read_product_amacd(*args)
    if file_type == FileType.AM__CTH_2B:
        return read_product_amcth(*args)
    else:
        return None


def _read_product(
    filepath: str,
    trim_to_frame: bool = True,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> Dataset:
    """Loads an EarthCARE product file as an `xarray.Dataset`.

    Args:
        filepath (str): Path to the product file.
        trim_to_frame (bool, optional): Whether to trim the dataset to latitude frame bounds. Defaults to True.
        modify (bool): If True, default modifications to the opened dataset will be applied
            (e.g., renaming dimension corresponding to height to "vertical"). Defaults to True.
        header (bool): If True, all header data will be included in the dataframe. Defaults to False.
        meta (bool): If True, select meta data from header (like orbit number and frame ID) will be included in the dataframe. Defaults to True.

    Returns:
        xarray.Dataset: Loaded (and optionally trimmed) dataset.
    """
    if not os.path.exists(filepath):
        raise FileNotFoundError(f"No product found under path: {filepath}")

    file_type = FileType.from_input(filepath)

    args: list = [
        filepath,
        file_type,
        modify,
        False,  # header will be read later
        False,  # meta data will be read later
    ]

    ds = _read_level1_product(*args)
    if not isinstance(ds, Dataset):
        ds = _read_level2a_product(*args)
    if not isinstance(ds, Dataset):
        ds = _read_level2b_product(*args)
    if not isinstance(ds, Dataset):
        trim_to_frame = False
        ds = _read_auxiliary_product(*args)
    if not isinstance(ds, Dataset):
        raise NotImplementedError(f"Product '{file_type}' not yet supported.")

    if modify and trim_to_frame:
        ds = trim_to_latitude_frame_bounds(ds)

    ds = add_header_and_meta_data(filepath, ds, header, meta)

    return ds


def read_product(
    input: str | Dataset,
    trim_to_frame: bool = True,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    in_memory: bool = False,
) -> Dataset:
    """Returns an `xarray.Dataset` from a Dataset or EarthCARE file path, optionally loaded into memory.

    Args:
        input (str or xarray.Dataset): Path to a EarthCARE file. If a `xarray.Dataset` is given it will be returned as is.
        trim_to_frame (bool, optional): Whether to trim the dataset to latitude frame bounds. Defaults to True.
        modify (bool): If True, default modifications to the opened dataset will be applied
            (e.g., renaming dimension corresponding to height to "vertical"). Defaults to True.
        header (bool): If True, all header data will be included in the dataframe. Defaults to False.
        meta (bool): If True, select meta data from header (like orbit number and frame ID) will be included in the dataframe. Defaults to True.
        in_memory (bool, optional): If True, ensures the dataset is fully loaded into memory. Defaults to False.

    Returns:
        xarray.Dataset: The resulting dataset.

    Raises:
        TypeError: If input is not a Dataset or string.
    """
    ds: Dataset
    if isinstance(input, Dataset):
        ds = input
    elif isinstance(input, str):
        if in_memory:
            with _read_product(
                filepath=input,
                trim_to_frame=trim_to_frame,
                modify=modify,
                header=header,
                meta=meta,
            ) as ds:
                ds = ds.load()
        else:
            ds = _read_product(
                filepath=input,
                trim_to_frame=trim_to_frame,
                modify=modify,
                header=header,
                meta=meta,
            )
    else:
        raise TypeError(
            f"Invalid input type! Expecting a opened EarthCARE dataset (xarray.Dataset) or a path to a EarthCARE product."
        )
    return ds



from typing import Final

# WGS 84
SEMI_MAJOR_AXIS_METERS: Final[float] = 6378137.0
SEMI_MINOR_AXIS_METERS: Final[float] = 6356752.314245

MEAN_EARTH_RADIUS_METERS = (SEMI_MAJOR_AXIS_METERS * 2 + SEMI_MINOR_AXIS_METERS) / 3



from logging import Logger
from typing import Final

import numpy as np
import pandas as pd

from ._exceptions import InvalidInputError
from ._parse_frame_ids import get_validated_frame_id
from ._parse_orbit_numbers import get_validated_orbit_number
from ._types import FrameIDStr, OrbitFrameStr, OrbitInt, _OrbitAndFrames

_FRAMES: Final[str] = "ABCDEFGH"
_NUM_FRAMES: Final[int] = 8


def get_validated_orbit_and_frame(
    orbit_and_frame: OrbitFrameStr, logger: Logger | None = None
) -> tuple[OrbitInt, FrameIDStr]:
    """Extracts validated orbit number and frame ID from string and raises InvalidInputError if string does not describe a orbit and frame"""
    try:
        orbit_number = get_validated_orbit_number(int(orbit_and_frame[0:-1]))
        frame_id = get_validated_frame_id(orbit_and_frame[-1])
    except Exception as e:
        exception_msg = f"{orbit_and_frame} is not a valid orbit and frame name. Valid names contain the orbit number followed by the frame id letter (e.g. 3000B or 03000B)."
        if logger:
            logger.exception(exception_msg)
        raise
    return orbit_number, frame_id


def get_complete_and_incomplete_orbits(
    orbit_and_frames: list[tuple[OrbitInt, OrbitFrameStr]] | None,
) -> tuple[list[OrbitInt], dict[OrbitFrameStr, list[OrbitInt]]]:
    """
    Finds complete orbits (i.e. where all frames are given) and incomplete orbits based on the given list of tuples.

    Args:
        orbit_and_frames: A list of tuples where each tuple contains
                          an orbit number (int) and a frame ID (str).

    Returns:
        tuple:
        - A list of complete orbits.
        - A dictionary where each key is a frame ID, and the value is a list of orbits assigned to that frame.
    """
    if not isinstance(orbit_and_frames, list):
        return [], {}
    if len(orbit_and_frames) == 0:
        return [], {}

    orbit_numbers: list[OrbitInt] = [oaf[0] for oaf in orbit_and_frames]
    frame_ids: list[OrbitFrameStr] = [oaf[1] for oaf in orbit_and_frames]

    df = pd.DataFrame(dict(orbit_number=orbit_numbers, frame_id=frame_ids))

    df_frames_per_orbit_lookup = df.groupby("orbit_number", as_index=False).agg(
        {"frame_id": lambda x: "".join(sorted("".join(x)))}
    )
    mask_complete_orbits = df_frames_per_orbit_lookup["frame_id"] == "ABCDEFGH"
    complete_orbits = df_frames_per_orbit_lookup.loc[mask_complete_orbits][
        "orbit_number"
    ].tolist()
    incomplete_orbits = df_frames_per_orbit_lookup.loc[~mask_complete_orbits][
        "orbit_number"
    ].tolist()
    df_incomplete_orbits = df.loc[df["orbit_number"].isin(incomplete_orbits)]
    df_orbits_per_frame_lookup = df_incomplete_orbits.groupby("frame_id").agg(
        {"orbit_number": list}
    )
    incomplete_orbits_frame_map = df_orbits_per_frame_lookup.to_dict()["orbit_number"]

    return complete_orbits, incomplete_orbits_frame_map


def get_frame_range(start_frame_id: str, end_frame_id: str) -> list[str]:
    """Returns list of frames in order of selected range (e.g. A-D -> ABCD and D-A -> DEFGHA)."""
    start_idx = _FRAMES.index(start_frame_id)
    end_idx = _FRAMES.index(end_frame_id)
    if end_idx < start_idx:
        end_idx = end_idx + _NUM_FRAMES
    frame_id_range = [
        _FRAMES[idx % _NUM_FRAMES] for idx in np.arange(start_idx, end_idx + 1)
    ]
    return frame_id_range


def parse_orbit_and_frames(
    orbit_and_frames: list[OrbitFrameStr] | None,
    start_orbit_and_frame: OrbitFrameStr | None,
    end_orbit_and_frame: OrbitFrameStr | None,
    logger: Logger | None = None,
) -> _OrbitAndFrames:
    orbit_and_frame_list: list[tuple[OrbitInt, FrameIDStr]] = []
    if isinstance(orbit_and_frames, list):
        orbit_and_frame_list = [
            get_validated_orbit_and_frame(oaf, logger=logger)
            for oaf in orbit_and_frames
        ]

    soaf: tuple[OrbitInt, FrameIDStr] | None = None
    eoaf: tuple[OrbitInt, FrameIDStr] | None = None
    if isinstance(start_orbit_and_frame, str):
        soaf = get_validated_orbit_and_frame(start_orbit_and_frame, logger=logger)
    if isinstance(end_orbit_and_frame, str):
        eoaf = get_validated_orbit_and_frame(end_orbit_and_frame, logger=logger)

    lower: OrbitInt | None = None
    upper: OrbitInt | None = None

    if isinstance(soaf, tuple) and isinstance(eoaf, tuple):
        so, sf = soaf
        eo, ef = eoaf
        if so > eo:
            raise InvalidInputError(
                f"End orbit and frame ({end_orbit_and_frame}) is smaller than start ({start_orbit_and_frame}) but needs to be greater or equal."
            )
        if so == eo:
            if sf > ef:
                raise InvalidInputError(
                    f"End orbit and frame ({end_orbit_and_frame}) is smaller than start ({start_orbit_and_frame}) but needs to be greater or equal."
                )
            _frame_range = get_frame_range(sf, ef)
            orbit_and_frame_list = orbit_and_frame_list + [
                (so, f) for f in _frame_range
            ]
        else:
            _frame_range = get_frame_range(sf, _FRAMES[-1])
            orbit_and_frame_list = orbit_and_frame_list + [
                (so, f) for f in _frame_range
            ]
            _frame_range = get_frame_range(_FRAMES[0], ef)
            orbit_and_frame_list = orbit_and_frame_list + [
                (eo, f) for f in _frame_range
            ]
            lower = so + 1
            upper = eo - 1
    elif isinstance(soaf, tuple):
        so, sf = soaf
        if sf == _FRAMES[0]:
            lower = so
        else:
            _frame_range = get_frame_range(sf, _FRAMES[-1])
            orbit_and_frame_list = orbit_and_frame_list + [
                (so, f) for f in _frame_range
            ]
            lower = so + 1
    elif isinstance(eoaf, tuple):
        eo, ef = eoaf
        if ef == _FRAMES[-1]:
            upper = eo
        else:
            _frame_range = get_frame_range(_FRAMES[0], ef)
            orbit_and_frame_list = orbit_and_frame_list + [
                (eo, f) for f in _frame_range
            ]
            upper = eo - 1

    orbit_and_frame_list = sorted(list(set(orbit_and_frame_list)))

    complete_orbits, incomplete_orbits_frame_map = get_complete_and_incomplete_orbits(
        orbit_and_frame_list
    )

    _complete_orbits = np.sort(complete_orbits)
    mask = np.full(_complete_orbits.shape, True, dtype=bool)

    if isinstance(lower, OrbitInt) and isinstance(upper, OrbitInt):
        for i, co in enumerate(_complete_orbits):
            if lower <= co <= upper:
                mask[i] = False
            elif co - 1 == upper:
                upper = upper + 1
                mask[i] = False
            else:
                break
    elif isinstance(lower, OrbitInt):
        for i, co in enumerate(_complete_orbits):
            if lower <= co:
                mask[i] = False
            else:
                break
    elif isinstance(upper, OrbitInt):
        for i, co in enumerate(_complete_orbits):
            if co <= upper:
                mask[i] = False
            elif co - 1 == upper:
                upper = upper + 1
                mask[i] = False
            else:
                break

    if isinstance(lower, OrbitInt):
        for i in range(len(_complete_orbits) - 1, -1, -1):
            co = _complete_orbits[i]
            if co + 1 == lower:
                lower = lower - 1
                mask[i] = False
            else:
                break

    assert _complete_orbits.shape == mask.shape

    _complete_orbits = _complete_orbits[mask].tolist()
    full_orbit_range: tuple[OrbitInt | None, OrbitInt | None] = (lower, upper)

    return _OrbitAndFrames(
        full_orbit_range=full_orbit_range,
        full_orbit_list=[int(o) for o in _complete_orbits],
        frame_orbits=incomplete_orbits_frame_map,
    )



import os
from enum import StrEnum
from typing import overload

import xarray as xr

from ..header_group import read_header_data
from ._type_alias import _format_file_type_string as format_file_type_string
from .file_info import FileInfoEnum

_short_hand_map = dict(
    # Level 1
    ATL_NOM_1B="A-NOM",
    ATL_DCC_1B="A-DCC",
    ATL_CSC_1B="A-CSC",
    ATL_FSC_1B="A-FSC",
    MSI_NOM_1B="M-NOM",
    MSI_BBS_1B="M-BBS",
    MSI_SD1_1B="M-SD1",
    MSI_SD2_1B="M-SD2",
    MSI_RGR_1C="M-RGR",
    BBR_NOM_1B="B-NOM",
    BBR_SNG_1B="B-SNG",
    BBR_SOL_1B="B-SOL",
    BBR_LIN_1B="B-LIN",
    CPR_NOM_1B="C-NOM",  # JAXA product
    # Level 2a
    ATL_FM__2A="A-FM",
    ATL_AER_2A="A-AER",
    ATL_ICE_2A="A-ICE",
    ATL_TC__2A="A-TC",
    ATL_EBD_2A="A-EBD",
    ATL_CTH_2A="A-CTH",
    ATL_ALD_2A="A-ALD",
    MSI_CM__2A="M-CM",
    MSI_COP_2A="M-COP",
    MSI_AOT_2A="M-AOT",
    CPR_FMR_2A="C-FMR",
    CPR_CD__2A="C-CD",
    CPR_TC__2A="C-TC",
    CPR_CLD_2A="C-CLD",
    CPR_APC_2A="C-APC",
    ATL_CLA_2A="A-CLA",  # JAXA product
    MSI_CLP_2A="M-CLP",  # JAXA product
    CPR_ECO_2A="C-ECO",  # JAXA product
    CPR_CLP_2A="C-CLP",  # JAXA product
    # Level 2b
    AM__MO__2B="AM-MO",
    AM__CTH_2B="AM-CTH",
    AM__ACD_2B="AM-ACD",
    AC__TC__2B="AC-TC",
    BM__RAD_2B="BM-RAD",
    BMA_FLX_2B="BMA-FLX",
    ACM_CAP_2B="ACM-CAP",
    ACM_COM_2B="ACM-COM",
    ACM_RT__2B="ACM-RT",
    ALL_DF__2B="ALL-DF",
    ALL_3D__2B="ALL-3D",
    AC__CLP_2B="AC-CLP",  # JAXA product
    ACM_CLP_2B="ACM-CLP",  # JAXA product
    ALL_RAD_2B="ALL-RAD",  # JAXA product
    # Auxiliary data
    AUX_MET_1D="X-MET",
    AUX_JSG_1D="X-JSG",
    # Orbit data
    MPL_ORBSCT="MPL-ORBSCT",
    AUX_ORBPRE="X-ORBPRE",
    AUX_ORBRES="X-ORBRES",
)


class FileType(FileInfoEnum):
    # Level 1
    ATL_NOM_1B = "ATL_NOM_1B"
    ATL_DCC_1B = "ATL_DCC_1B"
    ATL_CSC_1B = "ATL_CSC_1B"
    ATL_FSC_1B = "ATL_FSC_1B"
    MSI_NOM_1B = "MSI_NOM_1B"
    MSI_BBS_1B = "MSI_BBS_1B"
    MSI_SD1_1B = "MSI_SD1_1B"
    MSI_SD2_1B = "MSI_SD2_1B"
    MSI_RGR_1C = "MSI_RGR_1C"
    BBR_NOM_1B = "BBR_NOM_1B"
    BBR_SNG_1B = "BBR_SNG_1B"
    BBR_SOL_1B = "BBR_SOL_1B"
    BBR_LIN_1B = "BBR_LIN_1B"
    CPR_NOM_1B = "CPR_NOM_1B"  # JAXA product
    # Level 2a
    ATL_FM__2A = "ATL_FM__2A"
    ATL_AER_2A = "ATL_AER_2A"
    ATL_ICE_2A = "ATL_ICE_2A"
    ATL_TC__2A = "ATL_TC__2A"
    ATL_EBD_2A = "ATL_EBD_2A"
    ATL_CTH_2A = "ATL_CTH_2A"
    ATL_ALD_2A = "ATL_ALD_2A"
    MSI_CM__2A = "MSI_CM__2A"
    MSI_COP_2A = "MSI_COP_2A"
    MSI_AOT_2A = "MSI_AOT_2A"
    CPR_FMR_2A = "CPR_FMR_2A"
    CPR_CD__2A = "CPR_CD__2A"
    CPR_TC__2A = "CPR_TC__2A"
    CPR_CLD_2A = "CPR_CLD_2A"
    CPR_APC_2A = "CPR_APC_2A"
    ATL_CLA_2A = "ATL_CLA_2A"  # JAXA product
    MSI_CLP_2A = "MSI_CLP_2A"  # JAXA product
    CPR_ECO_2A = "CPR_ECO_2A"  # JAXA product
    CPR_CLP_2A = "CPR_CLP_2A"  # JAXA product
    # Level 2b
    AM__MO__2B = "AM__MO__2B"
    AM__CTH_2B = "AM__CTH_2B"
    AM__ACD_2B = "AM__ACD_2B"
    AC__TC__2B = "AC__TC__2B"
    BM__RAD_2B = "BM__RAD_2B"
    BMA_FLX_2B = "BMA_FLX_2B"
    ACM_CAP_2B = "ACM_CAP_2B"
    ACM_COM_2B = "ACM_COM_2B"
    ACM_RT__2B = "ACM_RT__2B"
    ALL_DF__2B = "ALL_DF__2B"
    ALL_3D__2B = "ALL_3D__2B"
    AC__CLP_2B = "AC__CLP_2B"  # JAXA product
    ACM_CLP_2B = "ACM_CLP_2B"  # JAXA product
    ALL_RAD_2B = "ALL_RAD_2B"  # JAXA product
    # Auxiliary data
    AUX_MET_1D = "AUX_MET_1D"
    AUX_JSG_1D = "AUX_JSG_1D"
    # Orbit data
    MPL_ORBSCT = "MPL_ORBSCT"
    AUX_ORBPRE = "AUX_ORBPRE"
    AUX_ORBRES = "AUX_ORBRES"

    @classmethod
    def from_input(cls, input: str | xr.Dataset) -> "FileType":
        """Infers the EarthCARE product type from a given file or dataset."""
        if isinstance(input, str):
            try:
                return cls[format_file_type_string(input)]
            except AttributeError:
                pass
            except KeyError:
                pass
            try:
                return cls(format_file_type_string(input))
            except ValueError:
                pass
            except KeyError:
                pass

        return get_file_type(input)

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    def to_shorthand(self, with_dash: bool = False):
        if with_dash:
            return _short_hand_map[self.value]
        else:
            return _short_hand_map[self.value].replace("-", "")


def _find_substring(target: str, substrings: list[str]) -> str:
    for sub in substrings:
        if sub in target:
            return sub
    raise ValueError(f"'{target}' does not contain a substring from '{substrings}'")


def _get_file_type_from_dataset(ds: xr.Dataset) -> FileType:
    try:
        file_type = str(ds.File_Type.values)
        return FileType(file_type)
    except (AttributeError, ValueError) as e:
        pass

    try:
        file_type = str(ds.file_type.values)
        return FileType(file_type)
    except (AttributeError, ValueError) as e:
        pass

    try:
        filename = os.path.basename(str(ds.filename.values)).rstrip(".h5")
        file_type = FileType.from_input(filename[9:19])
        return FileType(file_type)
    except (AttributeError, ValueError) as e:
        pass

    try:
        filepath = ds.encoding["source"]
        filename = os.path.basename(filepath)
        file_type = _find_substring(filename, FileType.list())
        return FileType(file_type)
    except (ValueError, KeyError) as e:
        raise ValueError(f"File name does not contain file type info. ({e})")


@overload
def get_file_type(product: str) -> FileType: ...
@overload
def get_file_type(product: xr.Dataset) -> FileType: ...
def get_file_type(product: str | xr.Dataset) -> FileType:
    if isinstance(product, str):
        with read_header_data(product) as ds:
            file_type = _get_file_type_from_dataset(ds)
    elif isinstance(product, xr.Dataset):
        file_type = _get_file_type_from_dataset(product)
    else:
        raise NotImplementedError()
    return file_type



import colorsys
from typing import List

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex

# def generate_hsl_circle_colormap(n: int, s: float = 0.7, l: float = 0.65) -> list[str]:
#     """
#     Generates `n` distinct colors evenly spaced around the HSL color wheel.

#     Args:
#         n (int): Number of colors to generate.
#         s (float, optional): Saturation component of the HSL color, between 0 and 1. Defaults to 0.65.
#         l (float, optional): Lightness component of the HSL color, between 0 and 1. Defaults to 0.5.

#     Returns:
#         list[str]: A list of hex color strings (e.g., "#ff0000").
#     """
#     colors: list[str] = []
#     for i in range(n):
#         h = i / n  # Hue: evenly spaced around the circle
#         r, g, b = colorsys.hls_to_rgb(h, l, s)
#         hex_color = "#{:02x}{:02x}{:02x}".format(
#             int(r * 255), int(g * 255), int(b * 255)
#         )
#         colors.append(hex_color)
#     return colors


def generate_hue_circle_colormap(n: int, rotation: float = 0.0) -> List[str]:
    """
    Generate `n` evenly spaced colors around the hue circle, similar to ggplot2's HCL palette,
    using standard Python + matplotlib.

    Args:
        n (int): Number of colors.
        rotation (float): Hue rotation in degrees (default 0.0).

    Returns:
        List[str]: List of hex color strings.
    """
    # These values approximate ggplot2 defaults (HCL)
    s = 0.65  # saturation
    l = 0.6  # lightness

    colors = []
    for i in range(n):
        hue = ((i / n) + rotation / 360.0) % 1.0
        r, g, b = colorsys.hls_to_rgb(hue, l, s)
        hex_color = "#{:02x}{:02x}{:02x}".format(
            int(r * 255), int(g * 255), int(b * 255)
        )
        colors.append(hex_color)
    return colors


def get_cmap(n=256, interpolate=False):
    name = "hsl"
    colors = generate_hue_circle_colormap(n)
    positions = np.linspace(0, 1, n)
    color_dict = list(zip(positions, colors))
    if interpolate:
        cmap = LinearSegmentedColormap.from_list(name, color_dict)
    else:
        cmap = ListedColormap(colors, name=name)
    return cmap



import numpy as np


def format_latitude(lat: float | int) -> str:
    """
    Format a latitude value as a string with degree and hemisphere (N/S).

    Returns an empty string if NaN. Raises `TypeError` for invalid input type.
    """
    if isinstance(lat, (float, int)):
        if np.isnan(lat):
            return ""
        elif lat < 0:
            return r"{:.1f}$^\circ$S".format(-1 * lat)
        else:
            return r"{:.1f}$^\circ$N".format(lat)
    raise TypeError(
        f"Given value `lat` hat wrong type '{type(lat).__name__}', expecting 'float' or 'int'"
    )


def format_longitude(lon: float | int) -> str:
    """
    Format a longitude value as a string with degree and direction (E/W).

    Returns an empty string if NaN. Raises `TypeError` for invalid input type.
    """
    if isinstance(lon, (float, int)):
        if np.isnan(lon):
            return ""
        elif lon < 0:
            return r"{:.1f}$^\circ$W".format(-1 * lon)
        else:
            return r"{:.1f}$^\circ$E".format(lon)
    raise TypeError(
        f"Given value `lon` hat wrong type '{type(lon).__name__}', expecting 'float' or 'int'"
    )


def format_coords(lat: float | int, lon: float | int, precision=4) -> str:
    """
    Format a pair of latitude and longitude values as a string coordinate string.

    Raises `TypeError` for invalid input type.
    """
    if not isinstance(lat, (float, int)):
        raise TypeError(
            f"Given value `lat` hat wrong type '{type(lat).__name__}', expecting 'float' or 'int'"
        )
    if not isinstance(lon, (float, int)):
        raise TypeError(
            f"Given value `lon` hat wrong type '{type(lon).__name__}', expecting 'float' or 'int'"
        )

    fstr = f"{{:.{precision}f}}"

    return f"({fstr.format(lat)}, {fstr.format(lon)})"



from logging import Logger
from typing import Final

from ._exceptions import InvalidInputError
from ._types import FrameIDStr

_FRAMES: Final[str] = "ABCDEFGH"
_NUM_FRAMES: Final[int] = 8


def get_validated_frame_id(
    frame_id: FrameIDStr,
    logger: Logger | None = None,
) -> FrameIDStr:
    """Formats frame ID and raises InvalidInputError if it is invalid"""
    try:
        frame_id = frame_id.upper()
        if len(frame_id) != 1:
            exception_msg = f"Got an empty string as frame ID. Valid frames are single letters from A to H."
            raise InvalidInputError(exception_msg)
        if frame_id not in "ABCDEFGH":
            exception_msg = f"{frame_id} is not a valid frame ID. Valid frames are single letters from A to H."
            raise InvalidInputError(exception_msg)
    except InvalidInputError as e:
        if logger:
            logger.exception(e)
        raise
    return frame_id


def parse_frame_ids(
    frame_ids: list[FrameIDStr] | None,
    logger: Logger | None = None,
) -> list[FrameIDStr]:
    if not isinstance(frame_ids, list):
        return []
    frame_ids = list(set([get_validated_frame_id(f, logger=logger) for f in frame_ids]))
    if len(frame_ids) == _NUM_FRAMES:
        return []
    else:
        return frame_ids



import re
import textwrap
from typing import Literal, TypeAlias

from ...utils.debug import get_calling_function_name

AxisInput: TypeAlias = Literal["x", "y", 0, 1]


def validate_axis_input(axis: AxisInput) -> Literal["x", "y"]:
    if isinstance(axis, int):
        axis = str(axis)
    elif not isinstance(axis, str):
        raise TypeError(
            f"{get_calling_function_name(2)}() for `axis` expected type '{str.__name__}' or '{int.__name__}' but got '{type(axis).__name__}' instead"
        )

    if axis == "0":
        axis = "x"
    elif axis == "1":
        axis = "y"

    if axis not in ["x", "y"]:
        raise ValueError(
            f"{get_calling_function_name(2)}() Invalid values given for `axis`: '{axis}' (expecting 'x', 'y' or respectively 0, 1)"
        )

    return axis


def wrap_label(label: str, width: int = 40) -> str:
    """Wrap a label string to a specified width, preserving units (in square brackets) and extra information.

    Args:
        label (str): The label string, optionally including units in square brackets.
        width (int, optional): Maximum width for each line. Defaults to 40.

    Returns:
        str: The wrapped label string.
    """
    wrapped_label = label
    match = re.match(r"([^\[]+)(\[[^\]]+\])?(.*)", label)
    if match:
        var_name = match.group(1).strip()
        units = match.group(2) or ""
        extra = match.group(3).strip()

        _width = width
        while len(var_name) % _width < _width / 2 and _width > 10:
            _width -= 1

        wrapped_var_name = textwrap.fill(var_name, width=_width)
        current = len(wrapped_var_name) % width

        if current + len(units) + len(extra) <= width:
            wrapped_label = f"{wrapped_var_name} {units} {extra}".strip()
        else:
            wrapped_label = f"{wrapped_var_name}\n{units} {extra}".strip()
    else:
        while len(label) % width < width / 2 and width > 10:
            width -= 1

        wrapped_label = textwrap.fill(label, width=width)
    return wrapped_label


def format_label(
    name: str, units: str | None = None, max_line_length: int | None = 40
) -> str:
    """Format a label with optional units and wrap it to a specified maximum line length.

    Args:
        name (str): The base name of the label.
        units (str | None, optional): The units to include in the label. Defaults to None.
        max_line_length (int | None, optional): The maximum length of each line. Defaults to 40.

    Returns:
        str: The formatted and wrapped label string.
    """
    if max_line_length is None:
        max_line_length = 40
    label = name
    if units is not None and (units != ""):
        label = f"{label} [{units}]"
    wrapped_label = wrap_label(label, max_line_length)
    return wrapped_label



import os
from pathlib import Path


def extend_filepath(filepath: str, suffix: str) -> str:
    """Appends a suffix to the filename before its extension.

    Args:
        filepath: Absolute file path.
        suffix: String to append to the filename.

    Returns:
        New file path with the suffix added.
    """

    p = Path(os.path.abspath(filepath))
    return str(p.with_name(f"{p.stem}{suffix}{p.suffix}"))



from ._concat import concat_datasets, read_products
from ._generic import read_product
from ._header_file import read_hdr_fixed_header
from ._search import search_product
from .auxiliary import rebin_xmet_to_vertical_track
from .file_info import get_product_infos
from .header_group import extract_basic_meta_data_from_header, read_header_data
from .science_group import read_science_data



from logging import Logger


def print_progress(
    msg: str,
    log_msg_prefix: str = "",
    line_length: int = 70,
    is_last: bool = False,
    logger: Logger | None = None,
):
    plt_msg: str = "Plotting ... "

    if is_last:
        plt_msg = ""

    _msg = f"{log_msg_prefix}{plt_msg}{msg}"
    num_spaces = line_length - len(_msg)
    if num_spaces < 0:
        _msg = f"{log_msg_prefix}{plt_msg}{msg[0 : num_spaces - 3]}..."
    else:
        _msg = f"{log_msg_prefix}{plt_msg}{msg}{' '*num_spaces}"

    if is_last:
        if logger:
            logger.info(_msg)
    else:
        print(_msg, end="\r")



from typing import Final, Literal

from ..color import Color
from .cmap import Cmap

STYLES: Final[list[Literal["qualitative", "intense", "dark", "light", "pastel"]]] = [
    "qualitative",
    "intense",
    "dark",
    "light",
    "pastel",
]


def get_cmap(
    style: Literal["qualitative", "intense", "dark", "light", "pastel"] = "intense",
    name: str | None = None,
) -> Cmap:
    if style not in STYLES:
        raise ValueError(f"""invalid style "{style}", valid styles are: {STYLES}""")

    colors = [
        "#E99590",
        "#D4A263",
        "#AEB050",
        "#79BB6B",
        "#22C198",
        "#00C0C3",
        "#59B5E2",
        "#A6A4EC",
        "#D793DF",
        "#EC8DBD",
    ]
    if style == "intense":
        colors = [
            "#F8766D",
            "#D89000",
            "#A3A500",
            "#39B600",
            "#00BF7D",
            "#00BFC4",
            "#00B0F6",
            "#9590FF",
            "#E76BF3",
            "#FF62BC",
        ]
    elif style == "dark":
        colors = [
            "#DC726B",
            "#C18500",
            "#959700",
            "#46A423",
            "#00AC77",
            "#00ABB0",
            "#009ED7",
            "#8A86E6",
            "#CA6AD4",
            "#E164A9",
        ]
    elif style == "light":
        colors = [
            "#FDB4B0",
            "#EABE8E",
            "#C9CA82",
            "#9ED494",
            "#71D9B7",
            "#63D7DA",
            "#8BCEF4",
            "#C3C1FD",
            "#ECB4F1",
            "#FFAED6",
        ]
    elif style == "pastel":
        colors = [
            "#FDC7C4",
            "#EECEAE",
            "#D6D7A7",
            "#B7DEB1",
            "#9EE1C9",
            "#9AE0E3",
            "#B0DAF6",
            "#D1D0FC",
            "#EEC7F3",
            "#FDC4DE",
        ]
    tmp_cmap = Cmap(colors, gradient=True)
    if not isinstance(name, str):
        name = f"HCL_{style}"
    n: int = 10
    final_cmap = Cmap(
        [Color(tmp_cmap(i / (n * 0.9)), is_normalized=True) for i in range(n)],
        name=name,
        gradient=True,
        circular=True,
    )
    return final_cmap


def get_cmaps() -> list[Cmap]:
    return [get_cmap("intense", name="HCL")] + [get_cmap(s) for s in STYLES]


def sample_hcl_colors(
    n: int,
    style: Literal["qualitative", "intense", "dark", "light", "pastel"] = "intense",
) -> list[Color]:
    cmap = get_cmap(style=style)
    return [Color(cmap(i / (n * 0.9)), is_normalized=True) for i in range(n)]


def sample_hcl_cmap(
    n: int,
    style: Literal["qualitative", "intense", "dark", "light", "pastel"] = "intense",
) -> Cmap:
    cmap = get_cmap(style=style)
    name = f"HCL_{style}_descrete"
    return Cmap(
        [Color(cmap(i / (n * 0.9)), is_normalized=True) for i in range(n)],
        name=name,
        gradient=False,
        circular=True,
    )



from dataclasses import dataclass
from logging import Logger

from ..utils._cli import get_counter_message
from ._eo_collection import EOCollection
from ._eo_product import (
    EOProduct,
    get_available_products,
    remove_duplicates_keeping_latest,
)


@dataclass
class EOSearchRequest:
    """This class contains all data required as input for the URL template of the OpenSearch API request to EO-CAT."""

    candidate_collections: list[EOCollection]
    product_type: str | None = None
    product_version: str | None = None
    radius: str | None = None
    lat: str | None = None
    lon: str | None = None
    bbox: str | None = None
    start_time: str | None = None
    end_time: str | None = None
    orbit_number: list[int] | None = None
    start_orbit_number: int | None = None
    end_orbit_number: int | None = None
    frame_id: str | None = None
    limit: int = 2000
    orbit_direction: str | None = None
    instrument: str | None = None
    product_id: str | None = None
    illumination_elevation_angle: str | None = None

    @property
    def stac_parameters(self) -> dict[str, str]:
        params: dict[str, str] = {}
        if isinstance(self.limit, int):
            params["limit"] = str(self.limit + 1)
        if isinstance(self.orbit_direction, str):
            params["orbitDirection"] = self.orbit_direction
        if isinstance(self.instrument, str):
            params["instrument"] = self.instrument
        if isinstance(self.product_id, str):
            params["uid"] = self.product_id
        if isinstance(self.illumination_elevation_angle, str):
            params["illuminationElevationAngle"] = self.illumination_elevation_angle
        # Product
        if isinstance(self.product_type, str):
            params["productType"] = self.product_type
        if isinstance(self.product_version, str):
            params["productVersion"] = self.product_version.lower()
        # Geo search
        if isinstance(self.radius, str):
            params["radius"] = self.radius
        if isinstance(self.lat, str):
            params["lat"] = self.lat
        if isinstance(self.lon, str):
            params["lon"] = self.lon
        if isinstance(self.bbox, str):
            params["bbox"] = self.bbox
        # Orbit number and frame ID
        if isinstance(self.orbit_number, list):
            orb_num_list_str = ",".join([str(int(o)) for o in self.orbit_number])
            params["orbitNumber"] = "{" + orb_num_list_str + "}"
        elif isinstance(self.start_orbit_number, int) and isinstance(
            self.end_orbit_number, int
        ):
            params["orbitNumber"] = (
                f"[{str(self.start_orbit_number)},{str(self.end_orbit_number)}]"
            )
        elif isinstance(self.start_orbit_number, int):
            params["orbitNumber"] = f"[{str(self.start_orbit_number)},99999]"
        elif isinstance(self.end_orbit_number, int):
            params["orbitNumber"] = f"[0,{str(self.end_orbit_number)}]"
        if isinstance(self.frame_id, str):
            params["frame"] = self.frame_id
        # Time search
        if isinstance(self.start_time, str) and isinstance(self.end_time, str):
            params["datetime"] = f"{self.start_time}/{self.end_time}"
        elif isinstance(self.start_time, str):
            params["datetime"] = f"{self.start_time}/"
        elif isinstance(self.end_time, str):
            params["datetime"] = f"/{self.end_time}"

        return params

    def run(
        self,
        logger: Logger | None = None,
        total_count: int | None = None,
        counter: int | None = None,
    ) -> list[EOProduct]:
        count_msg, _ = get_counter_message(counter=counter, total_count=total_count)

        if logger:
            logger.info(f"*{count_msg} {self.low_detail_summary}")
            logger.debug(f" {count_msg} {self}")

        if len(self.candidate_collections) == 0:
            if logger:
                logger.warning(
                    f" {count_msg} No collection was selected. Please make sure that you have added the appropriate collections for this product in the configuration file and that you are allowed to access to them."
                )
                return []

        _available_products: list[EOProduct] = []
        for cc in self.candidate_collections:
            _available_products = get_available_products(
                cc,
                params=self.stac_parameters,
                logger=logger,
            )

            _available_products = remove_duplicates_keeping_latest(_available_products)

            if len(_available_products) > 0:
                if logger:
                    logger.info(
                        f" {count_msg} Files found in collection '{cc.name}': {len(_available_products)}"
                    )
                    if len(_available_products) >= self.limit:
                        logger.warning(
                            f"The number of results equals the limit of results per search request ({self.limit}). Please divide your requests into smaller, manageable chunks (e.g., by selecting a shorter time range)."
                        )
                break

        return _available_products

    @property
    def low_detail_summary(self):
        msg = f"Search request: {self.product_type}"
        if self.product_version:
            msg = f"{msg}:{self.product_version}"
        if self.start_time and self.end_time:
            if self.start_time == self.end_time:
                msg = f"{msg}, time={self.start_time}"
            else:
                st_msg = self.start_time if self.start_time else "..."
                et_msg = self.end_time if self.end_time else "..."
                msg = f"{msg}, time_range=({st_msg},{et_msg})"
        if self.radius and self.lat and self.lon:
            msg = f"{msg}, radius=({self.radius}m, {self.lat}N, {self.lon}E)"
        if self.bbox:
            msg = f"{msg}, bbox={self.bbox}"
        if self.frame_id:
            msg = f"{msg}, frame={self.frame_id}"
        if self.start_orbit_number or self.end_orbit_number:
            so_msg = self.start_orbit_number if self.start_orbit_number else "..."
            eo_msg = self.end_orbit_number if self.end_orbit_number else "..."
            msg = f"{msg}, orbit_range=({so_msg},{eo_msg})"
        if self.orbit_number:
            o_msg = f"{self.orbit_number}"
            if len(o_msg.split(",")) > 6:
                o_msg = (
                    ",".join(o_msg.split(",")[0:2])
                    + f",... {len(o_msg.split(',')) - 4} more orbits ...,"
                    + ",".join(o_msg.split(",")[-2:])
                )
            msg = f"{msg}, orbits={o_msg.replace(',', ', ')}"
        return msg



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    ELEVATION_VAR,
    TROPOPAUSE_VAR,
    rename_common_dims_and_vars,
    rename_var_info,
)
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def read_product_atc(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens ATL_TC__2A file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="JSG_height",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="height",
        time_var="time",
        tropopause_var="tropopause_height",
        elevation_var="elevation",
    )

    ds = rename_var_info(
        ds,
        "simple_classification",
        "Simple classification",
        "Simple classification",
        "-",
    )
    for res, res_label in zip(
        ["", "_medium_resolution", "_low_resolution"], ["HiRes", "MedRes", "LowRes"]
    ):
        ds = rename_var_info(
            ds,
            f"classification{res}",
            name=f"Target classification ({res_label})",
            long_name=f"Target classification ({res_label})",
            units="-",
        )

    ds[TROPOPAUSE_VAR] = ds["tropopause_height"]
    ds[ELEVATION_VAR] = ds["elevation"]

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



def validate_baseline(baseline: str) -> str:
    if not isinstance(baseline, str):
        raise TypeError(f"Wrong type for baseline given ({type(baseline)}): {baseline}")

    if baseline == "latest":
        return ".."

    if len(baseline) != 2 or not baseline.isalpha():
        raise ValueError(f"Invalid product baseline: '{baseline}'")

    return baseline.upper()



import os
import warnings
from typing import Iterable, Literal

import matplotlib.patheffects as pe
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import font_manager
from matplotlib.axes import Axes
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.dates import date2num
from matplotlib.figure import Figure
from matplotlib.offsetbox import AnchoredText
from matplotlib.text import Text
from numpy.typing import ArrayLike, NDArray

from ...utils.constants import *
from ...utils.ground_sites import get_ground_site
from ...utils.np_array_utils import all_same
from ...utils.overpass import OverpassInfo, get_overpass_info
from ...utils.profile_data import (
    ProfileData,
    ensure_along_track_2d,
    ensure_vertical_2d,
    validate_profile_data_dimensions,
)
from ...utils.read import get_product_info, get_product_infos
from ...utils.read.product.file_info import FileType, ProductDataFrame
from ...utils.time import (
    TimeRangeLike,
    TimestampLike,
    format_time_range_text,
    to_timestamp,
    validate_time_range,
)
from ...utils.typing import DistanceRangeLike, ValueRangeLike
from ..color import Cmap, Color, ColorLike, get_cmap
from .along_track import AlongTrackAxisStyle, format_along_track_axis
from .colorbar import add_vertical_colorbar
from .defaults import get_default_cmap, get_default_norm, get_default_rolling_mean
from .format_strings import wrap_label
from .height_ticks import format_height_ticks


def add_text(
    ax: Axes,
    text: str,
    loc: str = "upper right",
    borderpad: float = 0,
    pad: float = 0.4,
    fontsize: str | float | None = None,
    fontweight: str | None = None,
    horizontalalignment: str = "left",
    color: Color | ColorLike | None = "black",
    is_shaded_text: bool = True,
    shade_linewidth: float = 3,
    shade_color: str = "white",
    shade_alpha: float = 0.8,
    is_box: bool = False,
    append_to: AnchoredText | str | None = None,
    zorder: int | float | None = None,
) -> AnchoredText:
    """
    Adds anchored text to a matplotlib Axes with optional shading and styling.

    Args:
        ax (matplotlib.axes.Axes): Target matplotlib Axes.
        text (str): Text string to display.
        loc (str): Anchor location in the Axes (e.g. 'upper right').
        borderpad (float): Padding between text and the border of the box.
        pad (float): Padding between box and the Axes.
        fontsize (str or float, optional): Font size of the text.
        fontweight (str, optional): Font weight (e.g. 'normal', 'bold').
        horizontalalignment (str): Horizontal alignment of the text.
        color (Color or ColorLike, optional): Text color.
        is_shaded_text (bool): If True, apply a white stroke around the text.
        shade_linewidth (float): Width of the stroke line.
        shade_color (str): Color of the stroke.
        shade_alpha (float): Opacity of the stroke.
        is_box (bool): If True, draw a box around the text.
        append_to (AnchoredText or str, optional): Extracts the given text string and adds the new text to it.
        zorder (int | float, optional): Drawing order of the plot element.

    Returns:
        AnchoredText: The text artist added to the Axes.
    """
    old_text: str | None = None
    if isinstance(append_to, AnchoredText):
        old_text = append_to.txt.get_text()
        append_to.remove()
    elif isinstance(append_to, str):
        old_text = append_to

    if isinstance(old_text, str):
        text = f"{old_text}{text}"

    path_effects = None
    if is_shaded_text:
        path_effects = [
            pe.withStroke(
                linewidth=shade_linewidth,
                foreground=shade_color,
                alpha=shade_alpha,
            )
        ]

    text_properties = {
        "size": fontsize,
        "fontweight": fontweight,
        "horizontalalignment": horizontalalignment,
        "path_effects": path_effects,
        "color": color,
    }

    anchored_text = AnchoredText(
        text,
        loc=loc,
        borderpad=borderpad,
        pad=pad,
        prop=text_properties,
        frameon=is_box,
        zorder=zorder,
    )
    ax.add_artist(anchored_text)
    return anchored_text


def add_title(
    ax: Axes,
    title: str,
    fontsize: str = "medium",
    loc: Literal["left", "center", "right"] | None = "center",
    **kwargs,
) -> Text:
    return ax.set_title(title, loc=loc, fontsize=fontsize, **kwargs)


# def get_earthcare_frame_string(ds: xr.Dataset) -> str:
#     if not hasattr(ds, "encoding"):
#         warnings.warn(
#             f"Dataset missing encoding attribute. Can not retrieve EarthCARE frame text"
#         )
#         text = ""
#     elif "source" in ds.encoding:
#         fp = os.path.basename(ds.encoding["source"])
#         df = get_product_info(fp, must_exist=False).to_dataframe()
#         text = f"{df.frame_id[0]}"
#     elif "sources" in ds.encoding:
#         df = get_product_infos(ds.encoding["sources"], must_exist=False)
#         orbit_start = str(df["orbit_number"].iloc[0]).zfill(5)
#         orbit_end = str(df["orbit_number"].iloc[-1]).zfill(5)
#         frame_start = df["frame_id"].iloc[0]
#         frame_end = df["frame_id"].iloc[-1]

#         if orbit_start == orbit_end:
#             text = (
#                 f"{orbit_start}{frame_start}"
#                 if frame_start == frame_end
#                 else f"{orbit_start}{frame_start}-{frame_end}"
#             )
#         else:
#             text = f"{orbit_start}{frame_start}-{orbit_end}{frame_end}"
#     else:
#         warnings.warn(
#             f"Dataset encoding does not contain source or sources. Can not retrieve EarthCARE frame text"
#         )
#         text = ""
#     return text


def get_earthcare_frame_string(data: xr.Dataset | ProductDataFrame) -> str:
    text: str = ""

    try:
        df: ProductDataFrame = get_product_infos(data)
    except ValueError as e:
        return text

    if len(df.shape) == 2 and df.shape[0] == 1:
        text = str(df["orbit_and_frame"][0])
    elif len(df.shape) == 2 and df.shape[0] > 1:
        orbit_start = str(df["orbit_number"].iloc[0]).zfill(5)
        orbit_end = str(df["orbit_number"].iloc[-1]).zfill(5)
        frame_start = df["frame_id"].iloc[0]
        frame_end = df["frame_id"].iloc[-1]

        if orbit_start == orbit_end:
            text = (
                f"{orbit_start}{frame_start}"
                if frame_start == frame_end
                else f"{orbit_start}{frame_start}-{frame_end}"
            )
        else:
            text = f"{orbit_start}{frame_start}-{orbit_end}{frame_end}"

    return text


def get_earthcare_file_type_baseline_string(data: xr.Dataset | ProductDataFrame) -> str:
    text: str = ""

    try:
        df: ProductDataFrame = get_product_infos(data)
    except ValueError as e:
        return text

    file_types = df["file_type"]
    baselines = df["baseline"]

    if not all_same(baselines):
        warnings.warn(f"The data contains multiple baselines: {baselines}")

    file_type: str = FileType.from_input(file_types[0]).to_shorthand()
    baseline: str = baselines[0]

    text = f"{file_type}:{baseline}"

    return text


def add_title_earthcare_frame(
    ax: Axes,
    ds: xr.Dataset,
    fontsize: str = "medium",
    loc: Literal["left", "center", "right"] | None = "right",
    color: Color | ColorLike | None = "black",
) -> Text:
    color = Color.from_optional(color)
    text = get_earthcare_frame_string(ds)
    return add_title(ax, text, fontsize=fontsize, loc=loc, color=color)
    # plt.title(text, loc=loc, fontsize=fontsize, color=color)


def add_title_earthcare_time(
    ax: Axes,
    ds: xr.Dataset | None = None,
    time_var: str = TIME_VAR,
    tmin: TimestampLike | None = None,
    tmax: TimestampLike | None = None,
    fontsize: str = "medium",
    loc: Literal["left", "center", "right"] | None = "left",
    color: Color | ColorLike | None = "black",
) -> Text:
    color = Color.from_optional(color)

    _tmin: TimestampLike | None = None
    _tmax: TimestampLike | None = None

    if isinstance(ds, xr.Dataset):
        _tmin = ds[time_var].values[0]
        _tmax = ds[time_var].values[-1]

    if isinstance(tmin, TimestampLike):
        _tmin = to_timestamp(tmin)

    if isinstance(tmax, TimestampLike):
        _tmax = to_timestamp(tmax)

    if isinstance(_tmin, TimestampLike) and isinstance(_tmax, TimestampLike):
        text = format_time_range_text(_tmin, _tmax)
    else:
        raise ValueError(
            f"Missing arguments. At least 'ds' or 'tmin' and 'tmax' must be given."
        )

    return add_title(ax, text, fontsize=fontsize, loc=loc, color=color)


def add_text_overpass_info(
    ax: Axes,
    info: OverpassInfo,
    zorder: int | float | None = 100,
) -> list[AnchoredText]:

    site = get_ground_site(info.site)
    site_name = site.name
    site_altitude = site.altitude
    site_coords = site.coordinates
    radius = info.site_radius_km
    samples = info.samples
    along_track_distance = info.along_track_distance_km
    closest_distance = info.closest_distance_km
    closest_time = info.closest_time

    _site_name = f" ({site_name})" if site_name != "" else ""
    _alt = f" {int(site_altitude)}m" if site_altitude is not None else ""
    _lat = "{:.3f}".format(site_coords[0]) + r"$^\circ\text{N}$"
    _lon = "{:.3f}".format(site_coords[1]) + r"$^\circ\text{E}$"
    _radius = f"Radius: {'{:.0f}'.format(radius)}km"
    _samples = f"Samples: {samples}"
    _along_track_distance = (
        ""
        if along_track_distance is None
        else f"\nAlong-track: {np.round(along_track_distance, decimals=3)}km"
    )
    _closest_distance = (
        ""
        if closest_distance is None
        else f"\nClosest: {np.round(closest_distance, decimals=3)}km at {pd.Timestamp(closest_time).strftime('%H:%M:%S')} UTC"
    )
    info_string = f"{_lat} {_lon}{_alt}{_site_name}\n{_radius}"
    add_text(
        ax,
        info_string,
        loc="upper left",
        horizontalalignment="left",
        fontsize="small",
        zorder=zorder,
    )
    info_string2 = f"{_samples}{_along_track_distance}{_closest_distance}"
    add_text(
        ax,
        info_string2,
        loc="lower left",
        horizontalalignment="left",
        fontsize="small",
        zorder=zorder,
    )
    text = f""

    t1 = add_text(
        ax,
        text,
        zorder=zorder,
    )

    return [t1]


def add_text_product_info(
    ax: Axes,
    ds: xr.Dataset,
    fontsize: str | float = "medium",
    loc: str = "upper right",
    color: Color | ColorLike | None = "black",
    append_to: AnchoredText | str | None = None,
    zorder: int | float | None = 100,
) -> AnchoredText:
    color = Color.from_optional(color)
    text_frame = get_earthcare_frame_string(ds)
    text_type_baseline = get_earthcare_file_type_baseline_string(ds)
    text = f"{text_frame}\n{text_type_baseline}"

    old_text: str | None = None
    if isinstance(append_to, AnchoredText):
        old_text = append_to.txt.get_text()
        append_to.remove()
    elif isinstance(append_to, str):
        old_text = append_to
    if isinstance(old_text, str):
        text = old_text
        if text_frame not in text:
            text = f"{text}\n{text_frame}"
        if text_type_baseline not in text:
            text = f"{text}\n{text_type_baseline}"

    horizontalalignment: str = "center"
    if "left" in loc:
        horizontalalignment = "left"
    elif "right" in loc:
        horizontalalignment = "right"

    return add_text(
        ax=ax,
        text=text,
        fontsize=fontsize,
        loc=loc,
        horizontalalignment=horizontalalignment,
        color=color,
        fontweight="bold",
        zorder=zorder,
    )


def format_var_label(
    name: str | None = None,
    units: str | None = None,
    da: xr.DataArray | None = None,
    label_len: int = 40,
) -> str:
    label: str

    if isinstance(da, xr.DataArray):
        if name is None and hasattr(da, "long_name"):
            name = da.long_name
        if units is None and hasattr(da, "units"):
            units = da.units

    if name is None:
        name = ""
    elif not isinstance(name, str):
        raise TypeError(
            f"Invalid type '{type(name).__name__}' for variable name: {name}. Expected type 'str'."
        )

    label = name

    if isinstance(units, str):
        if units == "":
            pass
        elif units.lower() not in ["-", "none"]:
            label = f"{name} [{units}]"

    label = wrap_label(label, label_len)

    return label



import os
import shutil
import time
import urllib.parse as urlp
from dataclasses import dataclass, field
from logging import Logger
from typing import Final

import pandas as pd
import requests
import requests.cookies

from ..utils import get_product_info
from ..utils._cli import console_exclusive_info, get_counter_message
from ._auth_oads import get_oads_authentification_cookies
from ._eo_collection import EOCollection
from ._eo_parameters import STACQueryParameter, get_available_parameters
from ._request import get_request_json, validate_request_response
from ._unzip import unzip_file

SUBDIR_NAME_AUX_FILES: Final[str] = "auxiliary_files"
SUBDIR_NAME_ORB_FILES: Final[str] = "orbit_files"
SUBDIR_NAME_L0__FILES: Final[str] = "level0"
SUBDIR_NAME_L1B_FILES: Final[str] = "level1b"
SUBDIR_NAME_L1C_FILES: Final[str] = "level1c"
SUBDIR_NAME_L2A_FILES: Final[str] = "level2a"
SUBDIR_NAME_L2B_FILES: Final[str] = "level2b"
MAX_DOWNLOAD_ATTEMPTS_PER_FILE: Final[int] = 3


def ensure_single_zip_extension(filename):
    """Returns given file name with a single .ZIP extension (e.g. 'file.ZIP.zip' -> 'file.ZIP')."""
    base_name, ext = os.path.splitext(filename)
    while ext.lower() == ".zip":
        base_name, ext = os.path.splitext(base_name)
    return base_name + ".ZIP"


def get_product_sub_dirname(product_name: str) -> str:
    """Returns level subfolder name of given product name."""
    if product_name in ["AUX_JSG_1D", "AUX_MET_1D"]:
        sub_dirname = SUBDIR_NAME_AUX_FILES
    elif product_name in ["MPL_ORBSCT", "AUX_ORBPRE", "AUX_ORBRES"]:
        sub_dirname = SUBDIR_NAME_ORB_FILES
    elif "0" in product_name.lower():
        sub_dirname = SUBDIR_NAME_L0__FILES
    elif "1b" in product_name.lower():
        sub_dirname = SUBDIR_NAME_L1B_FILES
    elif "1c" in product_name.lower():
        sub_dirname = SUBDIR_NAME_L1C_FILES
    elif "2a" in product_name.lower():
        sub_dirname = SUBDIR_NAME_L2A_FILES
    elif "2b" in product_name.lower():
        sub_dirname = SUBDIR_NAME_L2B_FILES
    return sub_dirname


def get_local_product_dirpath(dirpath_local, filename, create_subdirs=True):
    """Creates local path to file."""
    if create_subdirs:
        product_info = get_product_info(filename, must_exist=False)

        product_name = product_info.file_type.value
        year = str(product_info.start_sensing_time.year).zfill(4)
        month = str(product_info.start_sensing_time.month).zfill(2)
        day = str(product_info.start_sensing_time.day).zfill(2)

        sub_dirname = get_product_sub_dirname(product_name)
        product_dirpath_local = os.path.join(
            dirpath_local, sub_dirname, product_name, year, month, day
        )
    else:
        product_dirpath_local = dirpath_local
    return product_dirpath_local


@dataclass
class _DownloadResult:
    success: bool
    downloaded: bool
    unzipped: bool
    size_mb: float
    speed_mbs: float
    time: pd.Timedelta
    filepath: str


@dataclass(order=True)
class EOProduct:
    sort_index: tuple = field(init=False, repr=False)

    name: str
    server: str
    orbit_and_frame: str
    file_type: str
    version: str
    start_processing_time: pd.Timestamp
    url_download: str
    url_quicklook: str | None
    size: int | None

    def __post_init__(self):
        self.sort_index = (
            self.server,
            self.orbit_and_frame,
            self.start_processing_time,
        )

    def download(
        self,
        download_directory: str,
        is_overwrite: bool,
        is_unzip: bool,
        is_delete: bool,
        is_create_subdirs: bool,
        maap_token: str | None = None,
        oads_username: str | None = None,
        oads_password: str | None = None,
        oads_cookies_saml: requests.cookies.RequestsCookieJar | None = None,
        proxies: dict = {},
        counter: int | None = None,
        total_count: int | None = None,
        attempts: int = 3,
        chunk_size_bytes: int = 256 * 1024,
        logger: Logger | None = None,
    ) -> _DownloadResult:
        headers_maap: dict[str, str] | None = None
        if "maap" in self.url_download:
            if isinstance(maap_token, str):
                headers_maap = {"Authorization": "Bearer " + maap_token}
            else:
                raise ValueError(f"Download failed due to missing maap token")
        else:
            if not isinstance(oads_cookies_saml, requests.cookies.RequestsCookieJar):
                if not isinstance(oads_username, str) or not isinstance(
                    oads_password, str
                ):
                    raise ValueError(
                        f"Download failed due to missing oads username or password"
                    )
                oads_cookies_saml = get_oads_authentification_cookies(
                    dissemination_server=self.server,
                    username=oads_username,
                    password=oads_password,
                )

        _downloaded: bool = False
        _unzipped: bool = False
        _size_mb: float = 0.0
        _speed_mbs: float = 0.0
        _time: pd.Timedelta = pd.Timedelta(0.0)
        _filepath: str = "none"

        count_msg, _ = get_counter_message(counter=counter, total_count=total_count)

        _success: bool = True

        # Extracting the filename from the download link
        product_info = get_product_info(self.url_download, must_exist=False)
        file_name: str = product_info.name
        product_dirpath = get_local_product_dirpath(
            download_directory, file_name, create_subdirs=is_create_subdirs
        )

        # Make sure the local download_directory exists (if not create it)
        if not os.path.exists(product_dirpath):
            os.makedirs(product_dirpath)

        # Some files may be missing zip file extension so we need to fix them
        file_name = ensure_single_zip_extension(file_name)
        zip_file_path = os.path.join(product_dirpath, file_name)
        file_path = zip_file_path[0:-4]

        file_download_url = self.url_download

        for attempt in range(attempts):
            if attempt > 0:
                if logger:
                    logger.info(
                        f" {count_msg} Restarting (starting try {attempt + 1} of max. {MAX_DOWNLOAD_ATTEMPTS_PER_FILE})."
                    )

            # Check existing files
            zip_file_exists = os.path.exists(zip_file_path)
            file_exists = os.path.exists(file_path)

            # Decide if file will be downloaded and extracted
            try_download = is_overwrite or (not zip_file_exists and not file_exists)
            try_unzip = is_unzip and (is_overwrite or not file_exists)

            if not try_download:
                if is_unzip:
                    if logger:
                        logger.info(f" {count_msg} Skip file download.")
                else:
                    if logger:
                        logger.info(
                            f" {count_msg} Skip file download. (see <{zip_file_path}>)"
                        )
            if not try_unzip:
                if logger:
                    logger.info(f" {count_msg} Skip file unzip. (see <{file_path}>)")
            if not try_download and not try_unzip:
                break

            # Delete unnessecary zip files
            if is_delete and file_exists and zip_file_exists:
                os.remove(zip_file_path)
                zip_file_exists = False

            # Overwrite files
            if zip_file_exists and is_overwrite:
                os.remove(zip_file_path)
                zip_file_exists = False
            if file_exists and is_overwrite:
                shutil.rmtree(file_path)
                # os.remove(file_path)
                file_exists = False

            # Download zip file
            if try_download:
                try:
                    # Requesting the product download
                    if logger:
                        logger.debug(f" {count_msg} Requesting: {file_download_url}")

                    file_download_response: requests.Response
                    if headers_maap:
                        file_download_response = requests.get(
                            file_download_url,
                            headers=headers_maap,
                            stream=True,
                        )
                    else:
                        file_download_response = requests.get(
                            file_download_url,
                            cookies=oads_cookies_saml,
                            proxies=proxies,
                            stream=True,
                        )
                    validate_request_response(file_download_response, logger=logger)
                    
                    with open(zip_file_path, "wb") as f:
                        total_length_str = file_download_response.headers.get(
                            "content-length"
                        )
                        if isinstance(total_length_str, str):
                            self.size = int(total_length_str)
                        else:
                            total_length_str = file_download_response.headers.get(
                                "Content-Length"
                            )
                            if isinstance(total_length_str, str):
                                self.size = int(total_length_str)

                        # if self.size <= 0:
                        #     f.write(file_download_response.content)
                        # else:
                        current_length = 0
                        total_length = self.size
                        start_time = time.time()
                        progress_bar_length = 30
                        for data in file_download_response.iter_content(
                            chunk_size=chunk_size_bytes,
                        ):
                            current_length += len(data)
                            f.write(data)
                            done = int(
                                progress_bar_length * current_length / total_length
                            )
                            time_elapsed = time.time() - start_time
                            time_estimated = (
                                time_elapsed / current_length
                            ) * total_length
                            time_left = time.strftime(
                                "%H:%M:%S",
                                time.gmtime(int(time_estimated - time_elapsed)),
                            )
                            progress_bar = f"[{'#' * done}{'-' * (progress_bar_length - done)}]"
                            progress_percentage = f"{str(int((current_length / total_length) * 100)).rjust(3)}%"
                            elapsed_time = time.time() - start_time
                            _size_mb = current_length / 1024 / 1024
                            size_total = total_length / 1024 / 1024
                            _speed_mbs = (
                                _size_mb / elapsed_time if elapsed_time > 0 else 0
                            )  # MB/s
                            if logger:
                                if total_length > 0:
                                    console_exclusive_info(
                                        f"\r {count_msg} {progress_percentage} {progress_bar} {time_left} - {_speed_mbs:.2f} MB/s - {_size_mb:.2f}/{size_total:.2f} MB",
                                        end="\r",
                                    )
                                else:
                                    console_exclusive_info(
                                        f"\r {count_msg} Can not show progress estimate - {_speed_mbs:.2f} MB/s - {_size_mb:.2f}/? MB",
                                        end="\r",
                                    )
                        time_taken = time.strftime(
                            "%H:%M:%S",
                            time.gmtime(int(time.time() - start_time)),
                        )
                        _time = pd.Timedelta(time_taken)
                        if logger:
                            logger.info(
                                f" {count_msg} Download completed ({time_taken} - {_speed_mbs:.2f} MB/s - {_size_mb:.2f}/{size_total:.2f} MB)                   "
                            )
                except requests.exceptions.RequestException as e:
                    is_error_403_forbidden = False
                    if e.response is not None:  # Ensure response exists
                        is_error_403_forbidden = e.response.status_code == 403
                    if is_error_403_forbidden:
                        attempt = MAX_DOWNLOAD_ATTEMPTS_PER_FILE
                        if logger:
                            logger.error(f"DOWNLOAD FAILED: {e}")
                            logger.error(
                                f"Make sure that you only use OADS collections that you are allowed to access in your config.toml (see section 'Setup' in README)!"
                            )
                    else:
                        if logger:
                            logger.info(
                                f" {count_msg} DOWNLOAD FAILED for attempt {attempt + 1} of {MAX_DOWNLOAD_ATTEMPTS_PER_FILE}: {e}"
                            )
                        time.sleep(2)  # Wait for 2 seconds before retrying

                download_success = os.path.exists(zip_file_path)
                _downloaded = download_success
                _success &= download_success

            # Unzip zip file
            if try_unzip:
                _success = unzip_file(
                    zip_file_path,
                    delete=is_delete,
                    delete_on_error=True,
                    total_count=total_count,
                    counter=counter,
                    logger=logger,
                )
                unzip_success = os.path.exists(file_path)
                if unzip_success:
                    _unzipped = True
                _success &= unzip_success
            
            if _success:
                break

        return _DownloadResult(
            success=_success,
            downloaded=_downloaded,
            unzipped=_unzipped,
            size_mb=_size_mb,
            speed_mbs=_speed_mbs,
            time=_time,
            filepath=_filepath,
        )


def _create_search_url(
    collection: EOCollection,
    user_inputs: dict[str, str],
    logger: Logger | None = None,
) -> str:
    """Substitutes parameters given by the user into a search URL string if they match available parameters (else ignored)."""
    url_items = collection.url_items
    if not isinstance(url_items, str):
        return ""

    available_parameters = get_available_parameters(
        collection=collection,
        logger=logger,
    )

    available_parameter_dict = {eop.name: eop for eop in available_parameters}
    url_search = f"{url_items}?"
    for uik, uiv in user_inputs.items():
        p = available_parameter_dict.get(uik, None)
        if not isinstance(p, STACQueryParameter):
            continue

        if isinstance(p.enum, list) and uiv.lower() not in [e.lower() for e in p.enum]:
            continue

        url_search = f"{url_search}&{p.name}={uiv}"
    return url_search


def get_available_products(
    collection: EOCollection,
    params: dict[str, str],
    logger: Logger | None = None,
) -> list[EOProduct]:
    """Returns products matching user inputs from the specified collection."""
    url_search = _create_search_url(
        collection=collection,
        user_inputs=params,
        logger=logger,
    )

    if len(url_search) == 0:
        return []

    data = get_request_json(url=url_search, logger=logger)

    server: str
    url_download: str | None = None
    url_quicklook: str | None = None
    eo_products: list[EOProduct] = []
    for feature in data.get("features", []):
        assets = feature.get("assets")
        has_assets = isinstance(assets, dict)

        # eo-cat
        if has_assets:
            enclosure = assets.get("enclosure")
            if not isinstance(enclosure, dict):
                continue

            url_download = enclosure.get("href")
            if not isinstance(url_download, str):
                continue
            
            size = enclosure.get("file:size")

            server = str(urlp.urlparse(url_download).netloc)

            quicklook = assets.get("quicklook")
            if quicklook:
                url_quicklook = quicklook.get("href")
            else:
                url_quicklook = ""

            product_info = get_product_info(url_download, must_exist=False)

            eop = EOProduct(
                name=product_info.name.split(".")[0],
                server=server,
                orbit_and_frame=product_info.orbit_and_frame,
                file_type=product_info.file_type,
                version=product_info.baseline,
                start_processing_time=product_info.start_processing_time,
                url_download=url_download,
                url_quicklook=url_quicklook,
                size=size,
            )
            eo_products.append(eop)
            continue

        # maap
        size = feature.get("properties", {}).get("productInformation", {}).get("size", -1)
        for link in feature.get("links", []):
            server = ""
            url_download = None
            url_quicklook = ""

            _rel = link.get("rel")
            _type = link.get("type")

            if _rel in ["enclosure", "archives"] and _type == "application/zip":
                url_download = link.get("href", None)
                if not isinstance(url_download, str):
                    server = str(urlp.urlparse(url_download).netloc)
                    continue

            # TODO: Figure out where to find quicklooks on MAAP
            if _rel == "preview":
                url_quicklook = link.get("href", "")
            # url_quicklook = ""

            if isinstance(url_download, str) and isinstance(url_quicklook, str):
                product_info = get_product_info(url_download, must_exist=False)

                eop = EOProduct(
                    name=product_info.name,
                    server=server,
                    orbit_and_frame=product_info.orbit_and_frame,
                    file_type=product_info.file_type,
                    version=product_info.baseline,
                    start_processing_time=product_info.start_processing_time,
                    url_download=url_download,
                    url_quicklook="",
                    size=size,
                )
                eo_products.append(eop)
                break

    return eo_products


def remove_duplicates_keeping_latest(products: list[EOProduct]) -> list[EOProduct]:
    unique: dict[tuple[str, str], EOProduct] = {}

    for p in products:
        key = (p.file_type, p.orbit_and_frame)
        if (
            key not in unique
            or p.start_processing_time > unique[key].start_processing_time
        ):
            unique[key] = p

    return sorted(list(unique.values()))



import numpy as np
import xarray as xr
from matplotlib.colors import Colormap
from matplotlib.figure import Figure

from ...utils.constants import CM_AS_INCH, FIGURE_HEIGHT_CURTAIN, FIGURE_WIDTH_CURTAIN
from ...utils.read import read_product
from ...utils.time import to_timestamp
from ...utils.typing import ValueRangeLike
from ..color import ColorLike, get_cmap
from ..figure import MapFigure, create_column_subfigures


def ecswath(
    ds: xr.Dataset | str,
    var: str | None = None,
    n: int = 1,
    time_var: str = "time",
    style: str = "gray",
    border_color: ColorLike | None = "white",
    cmap: str | Colormap | None = "viridis",
    value_range: ValueRangeLike | None = None,
    show_colorbar: bool = True,
    track_color: ColorLike | None = "black",
    linewidth: float = 3.5,
    linestyle: str = "dashed",
    single_figsize: tuple[float, float] = (3, 8),
) -> tuple[Figure, list[MapFigure]]:
    ds = read_product(ds, in_memory=True)

    fig, axs = create_column_subfigures(n, single_figsize=single_figsize)
    tmin = to_timestamp(np.nanmin(ds[time_var].values))
    tmax = to_timestamp(np.nanmax(ds[time_var].values))
    tspan = (tmax - tmin) / n
    cb_alignment = "center"
    if n % 2 == 0:
        cb_alignment = "left"
    cb_height_ratio = "200%"
    if n == 1:
        cb_height_ratio = "100%"
    map_figs: list[MapFigure] = []
    for i in range(n):
        _show_colorbar = False
        if i == (n - 1) // 2:
            _show_colorbar = True
        show_text_time = False
        show_text_frame = False
        if i == 0:
            show_text_time = True
        if i == (n - 1):
            show_text_frame = True
        p = MapFigure(
            ax=axs[i],
            figsize=single_figsize,
            show_right_labels=False,
            show_top_labels=False,
            pad=0,
            style=style,
            show_text_time=show_text_time,
            show_text_frame=show_text_frame,
            border_color=border_color,
        )
        p = p.ecplot(
            ds,
            var,
            view="data",
            zoom_tmin=tmin + i * tspan,
            zoom_tmax=tmin + (i + 1) * tspan,
            colorbar=show_colorbar & _show_colorbar,
            cb_height_ratio=cb_height_ratio,
            cb_alignment=cb_alignment,
            cmap=get_cmap(cmap),
            value_range=value_range,
            color=track_color,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        p.grid_lines.ypadding = -3  # type: ignore
        p.grid_lines.ylabel_style = {  # type: ignore
            "color": "white",
            "fontsize": "small",
            "weight": "bold",
        }
        p.grid_lines.xpadding = -3  # type: ignore
        p.grid_lines.xlabel_style = {  # type: ignore
            "color": "white",
            "fontsize": "small",
            "weight": "bold",
        }

        map_figs.append(p)
    return fig, map_figs



import xarray as xr


def merge_datasets(
    ds1: xr.Dataset,
    ds2: xr.Dataset,
    keep_sec: bool = False,
) -> xr.Dataset:
    """Merges two datasets while keeping all global attributes from one dataset."""
    ds_merged = xr.merge([ds1, ds2])
    if keep_sec:
        ds_merged.attrs = ds2.attrs.copy()
        ds_merged.encoding = ds2.encoding.copy()
    else:
        ds_merged.attrs = ds1.attrs.copy()
        ds_merged.encoding = ds1.encoding.copy()
    return ds_merged



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap(
    alpha_clear: float = 1.0,
    alpha_missing: float = 1.0,
    alpha_attenuated: float = 1.0,
    alpha: float = 1.0,
):
    definitions = {
        -3: "Missing",
        -2: "Surface",
        -1: "Attenuated",
        0: "Clear",
        1: "Liquid cloud",
        2: "Ice cloud",
        3: "Aerosol",
        4: "Strat. cloud",
        5: "Strat. aerosol",
    }
    colors = [
        Color("#D3D3D3").set_alpha(alpha_missing),  # Missing
        Color("#000000").set_alpha(alpha),  # Surface
        Color("#696969").set_alpha(alpha_attenuated),  # Attenuated
        Color("#FFFFFF").set_alpha(alpha_clear),  # Clear
        Color("#0000FF").set_alpha(alpha),  # Liquid cloud
        Color("#00FFFF").set_alpha(alpha),  # Ice cloud
        Color("#DEB887").set_alpha(alpha),  # Aerosol
        Color("#800080").set_alpha(alpha),  # Stratospheric cloud
        Color("#FF00FF").set_alpha(alpha),  # Stratospheric aerosol
    ]
    cmap = Cmap(colors=colors, name="simple_classification").to_categorical(definitions)
    return cmap



import logging
import warnings
from typing import Iterable, Literal, Sequence

logger = logging.getLogger()

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import font_manager
from matplotlib.axes import Axes
from matplotlib.collections import PolyCollection
from matplotlib.colorbar import Colorbar
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.dates import date2num
from matplotlib.figure import Figure, SubFigure
from matplotlib.legend import Legend
from matplotlib.lines import Line2D
from matplotlib.offsetbox import AnchoredOffsetbox, AnchoredText
from matplotlib.text import Text
from numpy.typing import ArrayLike, NDArray

from ...utils.constants import *
from ...utils.ground_sites import GroundSite, get_ground_site
from ...utils.overpass import get_overpass_info
from ...utils.profile_data import (
    ProfileData,
    ensure_along_track_2d,
    ensure_vertical_2d,
    validate_profile_data_dimensions,
)
from ...utils.statistics import nan_max, nan_mean, nan_min, nan_sem, nan_std
from ...utils.time import (
    TimeRangeLike,
    TimestampLike,
    to_timestamp,
    to_timestamps,
    validate_time_range,
)
from ...utils.typing import DistanceRangeLike, ValueRangeLike, validate_numeric_range
from ..color import Cmap, Color, ColorLike, get_cmap
from ..save import save_plot
from .along_track import AlongTrackAxisStyle, format_along_track_axis
from .annotation import (
    add_text,
    add_text_product_info,
    add_title,
    add_title_earthcare_frame,
    format_var_label,
)
from .colorbar import add_vertical_colorbar
from .defaults import (
    get_default_cmap,
    get_default_norm,
    get_default_profile_range,
    get_default_rolling_mean,
)
from .height_ticks import format_height_ticks
from .ticks import format_numeric_ticks
from .value_range import select_value_range


def _convert_vertical_profile_to_step_function(
    values: ArrayLike, height: ArrayLike
) -> tuple[NDArray, NDArray]:
    values = np.asarray(values)
    height = np.asarray(height)

    hd1 = np.diff(height)
    hd2 = np.append(hd1[0], hd1)
    hd3 = np.append(hd1, hd1[-1])

    hnew1 = height - hd2 / 2
    hnew2 = height + hd3 / 2

    hnew = np.column_stack([hnew1, hnew2]).reshape(-1)
    vnew = np.repeat(values, 2)
    return vnew, hnew


def _highlight_height_range(
    ax: Axes,
    height_range: tuple[float, float],
    color: ColorLike | None = "gray",
    linewidth: float = 1,
    linestyle: str = "dashed",
    zorder: int | float | None = 0.9,
    alpha_fill: float = 0.15,
) -> None:
    _color = Color.from_optional(color)
    ax.axhspan(
        ymin=height_range[0],
        ymax=height_range[1],
        alpha=alpha_fill,
        color=_color,
        zorder=zorder,
        linewidth=0,
    )
    ax.axhline(
        y=height_range[0],
        color=_color,
        linestyle=linestyle,
        linewidth=linewidth,
        zorder=zorder,
    )
    ax.axhline(
        y=height_range[1],
        color=_color,
        linestyle=linestyle,
        linewidth=linewidth,
        zorder=zorder,
    )


class ProfileFigure:
    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] = (3, 4),
        dpi: int | None = None,
        title: str | None = None,
        height_axis: Literal["x", "y"] = "y",
        show_grid: bool = True,
        flip_height_axis: bool = False,
        show_legend: bool = False,
        show_height_ticks: bool = True,
        show_height_label: bool = True,
        height_range: DistanceRangeLike | None = None,
        value_range: ValueRangeLike | None = (0, None),
        label: str = "",
        units: str = "",
    ):
        self.fig: Figure
        if isinstance(ax, Axes):
            tmp = ax.get_figure()
            if not isinstance(tmp, (Figure, SubFigure)):
                raise ValueError(f"Invalid Figure")
            self.fig = tmp  # type: ignore
            self.ax = ax
        else:
            # self.fig: Figure = plt.figure(figsize=figsize, dpi=dpi)  # type: ignore
            # self.ax = self.fig.add_subplot()
            self.fig = plt.figure(figsize=figsize, dpi=dpi)
            self.ax = self.fig.add_axes((0.0, 0.0, 1.0, 1.0))
        self.title = title
        if isinstance(self.title, str):
            add_title(self.ax, title=self.title)
            # self.fig.suptitle(self.title)

        self.selection_time_range: tuple[pd.Timestamp, pd.Timestamp] | None = None
        self.info_text: AnchoredText | None = None

        self.ax_fill_between = (
            self.ax.fill_betweenx if height_axis == "y" else self.ax.fill_between
        )
        self.ax_set_hlim = self.ax.set_ylim if height_axis == "y" else self.ax.set_xlim
        self.ax_set_vlim = self.ax.set_ylim if height_axis == "x" else self.ax.set_xlim

        self.hmin: int | float | None = 0
        self.hmax: int | float | None = 40e3
        if isinstance(height_range, (Sequence, np.ndarray)):
            self.hmin = height_range[0]
            self.hmax = height_range[1]

        self.vmin: int | float | None = None
        self.vmax: int | float | None = None
        if isinstance(value_range, (Sequence, np.ndarray)):
            self.vmin = value_range[0]
            self.vmax = value_range[1]

        self.height_axis: Literal["x", "y"] = height_axis
        self.flip_height_axis = flip_height_axis
        self.value_axis: Literal["x", "y"] = "x" if height_axis == "y" else "y"

        self.show_grid: bool = show_grid

        self.label: str | None = label
        self.units: str | None = units

        self.ax_right: Axes | None = None
        self.ax_top: Axes | None = None

        self.show_legend: bool = show_legend
        self.legend_handles: list = []
        self.legend_labels: list[str] = []
        self.legend: Legend | None = None

        self.show_height_ticks: bool = show_height_ticks
        self.show_height_label: bool = show_height_label

        self._init_axes()

    def _init_axes(self) -> None:
        self.ax.grid(self.show_grid)

        self.ax_set_hlim(self.hmin, self.hmax)
        if self.vmin or self.vmax:
            if self.vmin and np.isnan(self.vmin):
                self.vmin = None
            if self.vmax and np.isnan(self.vmax):
                self.vmax = None
            self.ax_set_vlim(self.vmin, self.vmax)

        is_init = not isinstance(self.ax_right, Axes)

        if isinstance(self.ax_right, Axes):
            self.ax_right.remove()
        self.ax_right = self.ax.twinx()
        self.ax_right.set_ylim(self.ax.get_ylim())
        self.ax_right.set_yticklabels([])

        if isinstance(self.ax_top, Axes):
            self.ax_top.remove()
        self.ax_top = self.ax.twiny()
        self.ax_top.set_xlim(self.ax.get_xlim())
        format_numeric_ticks(
            self.ax_top,
            axis=self.value_axis,
            label=format_var_label(self.label, self.units),
            show_label=False,
        )
        self.ax_top.set_xticklabels([])

        if self.flip_height_axis:
            format_height_ticks(
                self.ax_right,
                axis=self.height_axis,
                show_tick_labels=self.show_height_ticks,
                label="Height" if self.show_height_label else None,
            )
            self.ax.set_yticklabels([])
        else:
            format_height_ticks(
                self.ax,
                axis=self.height_axis,
                show_tick_labels=self.show_height_ticks,
                label="Height" if self.show_height_label else None,
            )
        format_numeric_ticks(
            self.ax,
            axis=self.value_axis,
            label=format_var_label(self.label, self.units),
        )

        if self.show_legend and len(self.legend_handles) > 0:
            self.legend = self.ax.legend(
                handles=self.legend_handles,
                labels=self.legend_labels,
                fontsize="small",
                #   bbox_to_anchor=(1, 1),
                #   loc=2,
                bbox_to_anchor=(0, 1.015),
                loc="lower left",
                borderaxespad=0.25,
                edgecolor="white",
            )
        elif isinstance(self.legend, Legend):
            self.legend.remove()

    def plot(
        self,
        profiles: ProfileData | None = None,
        *,
        values: NDArray | None = None,
        time: NDArray | None = None,
        height: NDArray | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        error: NDArray | None = None,
        # Common args for wrappers
        label: str | None = None,
        units: str | None = None,
        value_range: ValueRangeLike | None = (0, None),
        height_range: DistanceRangeLike | None = None,
        time_range: TimeRangeLike | None = None,
        selection_height_range: DistanceRangeLike | None = None,
        show_mean: bool = True,
        show_std: bool = True,
        show_min: bool = False,
        show_max: bool = False,
        show_sem: bool = False,
        show_error: bool = False,
        color: str | ColorLike | None = None,
        alpha: float = 1.0,
        linestyle: str = "solid",
        linewidth: int | float = 1.5,
        ribbon_alpha: float = 0.2,
        show_grid: bool | None = None,
        zorder: int | float | None = 1,
        legend_label: str | None = None,
        show_legend: bool | None = None,
        show_steps: bool = DEFAULT_PROFILE_SHOW_STEPS,
    ) -> "ProfileFigure":
        color = Color.from_optional(color)

        if isinstance(show_legend, bool):
            self.show_legend = show_legend

        if isinstance(show_grid, bool):
            self.show_grid = show_grid
            self.ax.grid(self.show_grid)

        if isinstance(value_range, Iterable):
            if len(value_range) != 2:
                raise ValueError(
                    f"invalid `value_range`: {value_range}, expecting (vmin, vmax)"
                )
            else:
                if value_range[0] is not None:
                    self.vmin = value_range[0]
                if value_range[1] is not None:
                    self.vmax = value_range[1]
        else:
            value_range = (None, None)
        logger.debug(f"{value_range=}")

        if isinstance(profiles, ProfileData):
            values = profiles.values
            time = profiles.time
            height = profiles.height
            latitude = profiles.latitude
            longitude = profiles.longitude
            if not isinstance(label, str):
                label = profiles.label
            if not isinstance(units, str):
                units = profiles.units
            error = profiles.error
        elif values is None or height is None:
            raise ValueError(
                "Missing required arguments. Provide either a `VerticalProfiles` or all of `values` and `height`"
            )

        values = np.asarray(np.atleast_2d(values))
        if time is None:
            time = np.array([pd.Timestamp.now()] * values.shape[0])
        time = np.asarray(np.atleast_1d(time))
        height = np.asarray(height)
        if latitude is not None:
            latitude = np.asarray(latitude)
        if longitude is not None:
            longitude = np.asarray(longitude)

        vp = ProfileData(
            values=values,
            time=time,
            height=height,
            latitude=latitude,
            longitude=longitude,
            label=label,
            units=units,
            error=error,
        )

        vp.select_time_range(time_range)

        if isinstance(vp.label, str):
            self.label = vp.label
        if isinstance(vp.units, str):
            self.units = vp.units

        if height_range is not None:
            if isinstance(height_range, Iterable) and len(height_range) == 2:
                for i in [0, -1]:
                    height_range = list(height_range)
                    if height_range[i] is None:
                        height_range[i] = np.atleast_2d(vp.height)[0, i]
                    elif i == 0:
                        self.hmin = height_range[0]
                    elif i == -1:
                        self.hmax = height_range[-1]
                    height_range = tuple(height_range)
        else:
            height_range = (
                np.atleast_2d(vp.height)[0, 0],
                np.atleast_2d(vp.height)[0, -1],
            )

        if len(vp.height.shape) == 2 and vp.height.shape[0] == 1:
            h = vp.height[0]
        elif len(vp.height.shape) == 2:
            h = nan_mean(vp.height, axis=0)
        else:
            h = vp.height

        handle_mean: list[Line2D] | list[None] = [None]
        handle_min: list[Line2D] | list[None] = [None]
        handle_max: list[Line2D] | list[None] = [None]
        handle_std: PolyCollection | None = None
        handle_sem: PolyCollection | None = None

        if show_mean:
            if vp.values.shape[0] == 1:
                vmean = vp.values[0]
                show_std = False
                show_sem = False
                show_min = False
                show_max = False
            else:
                vmean = nan_mean(vp.values, axis=0)
            vnew, hnew = vmean, h
            if show_steps:
                vnew, hnew = _convert_vertical_profile_to_step_function(vmean, h)
            xy = (vnew, hnew) if self.height_axis == "y" else (hnew, vnew)
            handle_mean = self.ax.plot(
                *xy,
                color=color,
                alpha=alpha,
                zorder=zorder,
                linestyle=linestyle,
                linewidth=linewidth,
            )
            color = handle_mean[0].get_color()  # type: ignore

            value_range = select_value_range(vmean, value_range, pad_frac=0.01)
            if not (self.vmin is not None and self.vmin < value_range[0]):
                self.vmin = value_range[0]
            if not (self.vmax is not None and self.vmax > value_range[1]):
                self.vmax = value_range[1]

            if show_error and vp.error is not None:
                verror = vp.error.flatten()
                if show_steps:
                    verror, _ = _convert_vertical_profile_to_step_function(verror, h)
                handle_std = self.ax_fill_between(
                    hnew,
                    vnew - verror,
                    vnew + verror,
                    alpha=ribbon_alpha,
                    color=color,
                    linewidth=0,
                )

        if show_sem:
            vsem = nan_sem(vp.values, axis=0)
            if show_steps:
                vsem, _ = _convert_vertical_profile_to_step_function(vsem, h)
            handle_sem = self.ax_fill_between(
                hnew,
                vnew - vsem,
                vnew + vsem,
                alpha=ribbon_alpha,
                color=color,
                linewidth=0,
            )
        elif show_std:
            vstd = nan_std(vp.values, axis=0)
            if show_steps:
                vstd, _ = _convert_vertical_profile_to_step_function(vstd, h)
            handle_std = self.ax_fill_between(
                hnew,
                vnew - vstd,
                vnew + vstd,
                alpha=ribbon_alpha,
                color=color,
                linewidth=0,
            )

        if show_min:
            vmin = nan_min(vp.values, axis=0)
            vnew, hnew = vmin, h
            if show_steps:
                vnew, hnew = _convert_vertical_profile_to_step_function(vmin, h)
            xy = (vnew, hnew) if self.height_axis == "y" else (hnew, vnew)
            handle_min = self.ax.plot(
                *xy,
                color=color,
                alpha=alpha,
                zorder=zorder,
                linestyle="dashed",
                linewidth=linewidth,
            )
            color = handle_min[0].get_color()  # type: ignore

        if show_max:
            vmax = nan_max(vp.values, axis=0)
            vnew, hnew = vmax, h
            if show_steps:
                vnew, hnew = _convert_vertical_profile_to_step_function(vmax, h)
            xy = (vnew, hnew) if self.height_axis == "y" else (hnew, vnew)
            handle_max = self.ax.plot(
                *xy,
                color=color,
                alpha=alpha,
                zorder=zorder,
                linestyle="dashed",
                linewidth=linewidth,
            )
            color = handle_max[0].get_color()  # type: ignore

        # Legend labels
        if isinstance(legend_label, str):
            handle_std

            _handle: tuple | list = [
                *handle_mean,
                handle_std,
                handle_sem,
                *handle_min,
                *handle_max,
            ]
            _default_h = next(_h for _h in _handle if _h is not None)
            _handle = tuple([_h if _h is not None else _default_h for _h in _handle])
            self.legend_handles.append(_handle)
            self.legend_labels.append(legend_label)

        if selection_height_range:
            _shr: tuple[float, float] = validate_numeric_range(selection_height_range)
            _highlight_height_range(
                ax=self.ax,
                height_range=_shr,
            )

        self._init_axes()

        # format_height_ticks(self.ax, axis=self.height_axis)
        # format_numeric_ticks(self.ax, axis=self.value_axis, label=self.label)

        return self

    def ecplot(
        self,
        ds: xr.Dataset,
        var: str,
        *,
        time_var: str = TIME_VAR,
        height_var: str = HEIGHT_VAR,
        lat_var: str = TRACK_LAT_VAR,
        lon_var: str = TRACK_LON_VAR,
        error_var: str | None = None,
        along_track_dim: str = ALONG_TRACK_DIM,
        values: NDArray | None = None,
        time: NDArray | None = None,
        height: NDArray | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        error: NDArray | None = None,
        site: str | GroundSite | None = None,
        radius_km: float = 100.0,
        # Common args for wrappers
        value_range: ValueRangeLike | None = None,
        height_range: DistanceRangeLike | None = (0, 40e3),
        time_range: TimeRangeLike | None = None,
        selection_height_range: DistanceRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        zorder: int | float | None = 1,
        legend_label: str | None = "EarthCARE",
        show_legend: bool | None = None,
        show_steps: bool = DEFAULT_PROFILE_SHOW_STEPS,
        **kwargs,
    ) -> "ProfileFigure":
        # Collect all common args for wrapped plot function call
        local_args = locals()
        # Delete all args specific to this wrapper function
        del local_args["self"]
        del local_args["ds"]
        del local_args["var"]
        del local_args["time_var"]
        del local_args["height_var"]
        del local_args["lat_var"]
        del local_args["lon_var"]
        del local_args["error_var"]
        del local_args["along_track_dim"]
        del local_args["site"]
        del local_args["radius_km"]
        # Delete kwargs to then merge it with the residual common args
        del local_args["kwargs"]
        all_args = {**local_args, **kwargs}

        if all_args["values"] is None:
            all_args["values"] = ds[var].values
        if all_args["time"] is None:
            all_args["time"] = ds[time_var].values
        if all_args["height"] is None:
            all_args["height"] = ds[height_var].values
        if all_args["latitude"] is None:
            all_args["latitude"] = ds[lat_var].values
        if all_args["longitude"] is None:
            all_args["longitude"] = ds[lon_var].values
        if all_args["error"] is None and isinstance(error_var, str):
            all_args["error"] = ds[error_var].values

        # Set default values depending on variable name
        if label is None:
            all_args["label"] = (
                "Values" if not hasattr(ds[var], "long_name") else ds[var].long_name
            )
        if units is None:
            all_args["units"] = "-" if not hasattr(ds[var], "units") else ds[var].units
        if value_range is None:
            all_args["value_range"] = get_default_profile_range(var)

        self.plot(**all_args)

        return self

    def show(self):
        self.fig.tight_layout()
        self.fig.show()

    def save(self, filename: str = "", filepath: str | None = None, **kwargs):
        save_plot(fig=self.fig, filename=filename, filepath=filepath, **kwargs)
